# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''Defines various tools for wxPython.'''

from __future__ import division
import wx


# employ generic caching decorator here to make shorter

_background_color = None
def get_background_color():
    '''Get the default garlicsim_wx background color'''
    global _background_color
    if _background_color is not None:
        return _background_color
    
    result = wx.Color(212, 208, 200)
    '''
    NOTE I'M ACTUALLY USING A CONSTANT COLOR NOW. THIS COMMENT IS NOT RELEVANT
    RIGHT NOW:    
    todo: Not sure it's the right system color. Find the right one by comparing
    on different platforms. The right one is probably one of these:
    
    ['SYS_COLOUR_MENUBAR', 'SYS_COLOUR_SCROLLBAR', 'SYS_COLOUR_3DFACE',
    'SYS_COLOUR_INACTIVECAPTIONTEXT', 'SYS_COLOUR_3DLIGHT', 'SYS_COLOUR_MENU',
    'SYS_COLOUR_INACTIVEBORDER', 'SYS_COLOUR_BTNFACE',
    'SYS_COLOUR_ACTIVEBORDER']
    '''
    
    _background_color = result
    return result


_background_brush = None
def get_background_brush():
    '''Get the default garlicsim_wx background brush.'''
    global _background_brush
    if _background_brush is not None:
        return _background_brush
    result = wx.Brush(get_background_color())
    _background_brush = result
    return result


def post_event(evt_handler, event_binder, source=None):
    '''Post an event to an evt_handler.'''
    # todo: Use wherever I post events
    event = wx.PyEvent(source.GetId() if source else 0)
    event.SetEventType(event_binder.evtType[0])
    wx.PostEvent(evt_handler, event)
    
    
class Key(object):    
    '''A key combination.'''

    def __init__(self, key_code, cmd=False, alt=False, shift=False):

        self.key_code = key_code        
        '''The numerical code of the pressed key.'''
        
        self.cmd = cmd
        '''Flag saying whether the ctrl/cmd key was pressed.'''
        
        self.alt = alt
        '''Flag saying whether the alt key was pressed.'''
        
        self.shift = shift
        '''Flag saying whether the shift key was pressed.'''
        
        
    @staticmethod
    def get_from_key_event(event):
        '''Construct a Key from a wx.EVT_KEY_DOWN event.'''
        return Key(event.GetKeyCode(), event.CmdDown(),
                   event.AltDown(), event.ShiftDown())
    
    def __hash__(self):
        return hash(tuple(sorted(tuple(vars(self)))))
    
    def __eq__(self, other):
        if not isinstance(other, Key):
            return NotImplemented
        return self.key_code == other.key_code and \
               self.cmd == other.cmd and \
               self.shift == other.shift and \
               self.alt == other.alt
        
    
def iter_rects_of_region(region):
    '''Iterate over the rects of a region.'''
    i = wx.RegionIterator(region)
    while i.HaveRects():
        yield i.GetRect()
        i.Next()
        





# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This module defines the TreeBrowser class. See its documentation for more info.
'''
#todo: I think the refresh should be made more efficient


import os
from math import *

import pkg_resources
import wx
from wx.lib.scrolledpanel import ScrolledPanel

import garlicsim_wx.general_misc.vectorish as vectorish
from garlicsim_wx.general_misc import emitters
from garlicsim_wx.general_misc import wx_tools
from garlicsim_wx.general_misc.flag_raiser import FlagRaiser
import garlicsim
import garlicsim_wx
from garlicsim_wx.widgets import WorkspaceWidget

from . import images as __images_package
images_package = __images_package.__name__

connector_length = 10 # length of connecting line between elements


class TreeBrowser(ScrolledPanel, WorkspaceWidget):
    '''Widget for browsing a garlicsim.data_structures.Tree.'''
    def __init__(self, frame): # todo: on mouse drag should pause like seek bar does
        ScrolledPanel.__init__(self, frame, size=(100, 100),
                               style=wx.SUNKEN_BORDER)
        WorkspaceWidget.__init__(self, frame)
        
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        
        self.SetupScrolling()
        #self.SetScrollRate(20,20)
        #self.sizer=wx.BoxSizer(wx.VERTICAL)
        #self.panel=wx.StaticBitmap(self,-1,wx.Bitmap("images\\snail.gif", wx.BITMAP_TYPE_ANY))#wx.Panel(self,-1,size=(-1,200))#wx.TextCtrl(self, -1, size=(-1,200), style=wx.TE_MULTILINE)
        #self.panel=wx.Panel(self,-1,size=(-1,100))
        #self.sizer.Add(self.panel,1,wx.EXPAND)
        #self.SetSizer(self.sizer)
        #self.EnableScrolling(True, True)
        #self.SetScrollbars(5, 30, 1055, 40)
        #self.sizer.Fit(self)
        #self.Centre()
        #self.SetVirtualSize((1000,1000))

        self.Bind(wx.EVT_PAINT, self.on_paint)
        self.Bind(wx.EVT_SIZE, self.on_size)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.on_mouse_event)

        self.menu = garlicsim_wx.general_misc.cute_menu.CuteMenu.add_menus(
            [garlicsim_wx.misc.menu_bar.node_menu.NodeMenu(self.frame),
             garlicsim_wx.misc.menu_bar.block_menu.BlockMenu(self.frame)]
        )
        
        self.tree_remapping_flag = False
        self.recalculation_flag = False
        
        self.needs_tree_remapping_emitter = \
            self.gui_project.emitter_system.make_emitter(
                inputs=(
                    self.gui_project.tree_structure_modified_emitter,
                ),
                outputs=(
                    FlagRaiser(self, 'tree_remapping_flag'),
                ),
                name='needs_tree_remapping',
            )

        
        self.needs_recalculation_emitter = \
            self.gui_project.emitter_system.make_emitter(
                inputs=(
                    self.needs_tree_remapping_emitter,
                    self.gui_project.active_node_changed_or_modified_emitter,
                    self.gui_project.tree_modified_on_path_emitter,                
                    # Note that if there's a non-structure tree change not on
                    # the path it won't affect us.
                ),
                outputs=(
                    FlagRaiser(self, 'recalculation_flag'),
                ),
                name='needs_recalculation',
            )
        
        self.clickable_map = {}
        
        elements_raw = {            
            'Untouched': 'graysquare.png',
            'Touched': 'graystar.png',
            'Unfinalized Touched': 'grayunfinalizedstar.png',
            'Untouched End': 'grayendsquare.png',
            'Touched End': 'grayendstar.png',
            'Block': 'grayblock.png',
            'Active Untouched': 'orangesquare.png',
            'Active Touched': 'orangestar.png',
            'Active Unfinalized Touched': 'orangeunfinalizedstar.png',
            'Active Untouched End': 'orangeendsquare.png',
            'Active Touched End': 'orangeendstar.png',
            'Active Block': 'orangeblock.png',
        }
        
        self.elements = {}
        for key in elements_raw:
            stream = pkg_resources.resource_stream(images_package,
                                                      elements_raw[key])
            self.elements[key] = wx.BitmapFromImage(
                wx.ImageFromStream(
                    stream,
                    wx.BITMAP_TYPE_ANY
                )
            )

    def on_paint(self, event):
        '''Refresh the tree browser.'''

        event.Skip()
        
        # todo: optimize so it groks the tree only when needed.
        
        self.tree_remapping_flag = False
        self.recalculation_flag = False
        # todo: now we just lower these flags retardedly, in future there will
        # be __recalculate
        
        if self.gui_project is None or \
           self.gui_project.project.tree is None or \
           len(self.gui_project.project.tree.roots) == 0:
            
            return


        pen = wx.Pen('Black', 1, wx.SOLID)
        pen.SetCap(wx.CAP_PROJECTING)
        pen.SetJoin(wx.JOIN_ROUND)

        dc = NiftyPaintDC(self, self.gui_project,
                          self.CalcScrolledPosition((0, 0)), self)
        
        dc.SetBackground(wx_tools.get_background_brush())
        dc.Clear()
        
        (self.clickable_map, (width, height)) = \
            dc.draw_tree(self.gui_project.project.tree)
        
        self.SetVirtualSize((width,height))
        dc.Destroy()

    def on_size(self, e=None):
        self.Refresh()
        if e is not None:
            e.Skip()

    def on_mouse_event(self, e):
        #todo: should catch drag to outside of the window
        #(x,y)=self.CalcUnscrolledPosition(e.GetPositionTuple())

        (x, y) = e.GetPositionTuple()
        
        if e.LeftDClick():
            self.gui_project.toggle_playing()


        if e.LeftIsDown():
            thing = self.search_map(x,y)
            #print(thing)
            if thing is None:
                #maybe deselect?
                pass
            elif isinstance(thing, garlicsim.data_structures.End):
                self.gui_project.set_active_node(thing.parent)
            else:
                self.gui_project.set_active_node(thing)
                
        if e.RightDown():
            self.gui_project.stop_playing()
            thing = self.search_map(x,y)
            if thing is None:
                #Deselect!
                pass
            else:
                self.gui_project.set_active_node(thing)
                
            self.PopupMenu(self.menu, e.GetPosition())


    def search_map(self, x, y):
        for key in self.clickable_map:
            (a, b, c, d) = key
            if (a <= x <= c) and (b <= y <= d):
                thing = self.clickable_map[key]
                if isinstance(thing, garlicsim.data_structures.Block):
                    ratio = (x - a) / float(c - a)
                    index = int(round(ratio*(len(thing)-1)))
                    return thing[index]
                else:
                    return thing
        return None



class NiftyPaintDC(wx.BufferedPaintDC):
    '''A PaintDC used to paint the tree in a tree browser.'''
    
    def __init__(self, window, gui_project, origin, tree_browser, *args, **kwargs):
        wx.BufferedPaintDC.__init__(self, window, *args, **kwargs)
        self.gui_project = gui_project
        self.origin = origin
        self.tree_browser = tree_browser
        
        

    def draw_sub_tree(self, point, tree, start):
        make_block_stripe = False

        if isinstance(start, garlicsim.data_structures.Block):
            
            type = "Block"
            kids = start[-1].children
            if start == self.active_soft_block:
                make_block_stripe = True
                type = "Active " + type
                
        elif isinstance(start, garlicsim.data_structures.Node):
            
            kids = start.children
            
            if start.touched:
                type = "Touched"
            else:
                type = "Untouched"
                
            if start.still_in_editing:
                type = 'Unfinalized ' + type
                    
            if start == self.active_soft_block:
                type = "Active " + type
                
        else:
            
            raise Exception


        if make_block_stripe is True:

            bitmap = self.tree_browser.elements["Block"]
            self.DrawBitmapPoint(bitmap, point, useMask=True)
            bitmap_size = bitmap.GetSize()
            second_bitmap = self.tree_browser.elements[type]

            slice = [None, None]
            length = float(len(start))
            slice[0] = start.index(self.active_node) / length
            slice[1] = slice[0] + (1 / length)

            screen_slice = [
                floor(point[0] + 2 + (bitmap_size[0] - 4) * slice[0]),
                ceil(point[0] + 2 + (bitmap_size[0] - 4) * slice[1])
            ]
            region = wx.Region(screen_slice[0], point[1],
                               screen_slice[1] - screen_slice[0],
                               bitmap_size[1])
            
            self.SetClippingRegionAsRegion(region)
            self.DrawBitmapPoint(second_bitmap, point, useMask=True)
            self.DestroyClippingRegion()

        else:
            bitmap = self.tree_browser.elements[type]
            self.DrawBitmapPoint(bitmap, point, useMask=True)
            bitmap_size = bitmap.GetSize()

        temp = (point[0],
                point[1],
                point[0] + bitmap_size[0],
                point[1] + bitmap_size[1])
        
        self.clickable_map[temp] = start
        del temp



        last_height = 0
        total_height = 0
        self_width = bitmap_size[0] + connector_length
        max_width = self_width
        line_start = vectorish.add(
            point,
            (
                bitmap_size[0] - 1,
                bitmap_size[1] // 2
            )
        )
        
        for kid in kids:
            line_end = vectorish.add(
                point,
                (
                    self_width + 1,
                 total_height + bitmap_size[1] // 2
                )
            )
            
            temp = vectorish.add(point, (self_width, total_height))
            (new_width, new_height) = \
                self.draw_sub_tree(temp, tree, kid.soft_get_block())
            del temp
            self.DrawLinePoint(line_start, line_end)
            max_width = max(max_width, self_width + new_width)
            total_height += new_height
            
        
        ends = start.ends if isinstance(start, garlicsim.data_structures.Node) \
             else start[-1].ends
        for end in ends:
            line_end = vectorish.add(
                point,
                (
                    self_width + 1,
                 total_height + bitmap_size[1] // 2
                )
            )
            
            temp = vectorish.add(point, (self_width, total_height))
            (new_width, new_height) = \
                self.draw_end(temp, tree, end)
            del temp
            self.DrawLinePoint(line_start, line_end)
            max_width = max(max_width, self_width + new_width)
            total_height += new_height

        return (
            max_width,
            max(
                total_height,
                bitmap_size[1] + connector_length
            )
        )

    
    def draw_end(self, point, tree, start):

        assert isinstance(start, garlicsim.data_structures.End)
        
        bitmap = self.tree_browser.elements['Untouched End']
        self.DrawBitmapPoint(bitmap, point, useMask=True)
        bitmap_size = bitmap.GetSize()

        temp = (point[0],
                point[1],
                point[0] + bitmap_size[0],
                point[1] + bitmap_size[1])
        
        self.clickable_map[temp] = start
        
        last_height = 0
        total_height = 0
        self_width = bitmap_size[0] + connector_length
        max_width = self_width
        line_start = vectorish.add(
            point,
            (
                bitmap_size[0] - 1,
                bitmap_size[1] // 2
            )
        )
        
        

        return (
            max_width,
            max(
                total_height,
                bitmap_size[1] + connector_length
            )
        )
    
    def draw_tree(self, tree):
        '''
        Draw the tree.
        '''

        if self.gui_project:
            self.clickable_map = {}
            self.active_node = self.gui_project.active_node
            try:
                self.active_soft_block = self.active_node.soft_get_block()
            except AttributeError:
                self.active_soft_block = None
    
            sizes = []
            pos = vectorish.add((connector_length, connector_length), self.origin)
            for root in tree.roots:
                size = self.draw_sub_tree(
                    pos,
                    tree,
                    root.soft_get_block()
                )
                pos = vectorish.add(pos, (size[0], 0))
                sizes.append(size)
            
            width = sum(size[0] for size in sizes) + (connector_length * len(sizes))
            height = max(size[1] for size in sizes) + connector_length
            return (self.clickable_map, (width, height))


'''

Maybe I'll use this sometime:

                        "Selected Untouched": wx.Bitmap("images\\bluesquare.png", wx.BITMAP_TYPE_ANY),    \
                       "Selected Touched": wx.Bitmap("images\\bluestar.png", wx.BITMAP_TYPE_ANY),    \
                       "Selected Block": wx.Bitmap("images\\blueblock.png", wx.BITMAP_TYPE_ANY),    \
'''


'''

Maybe I'll use this sometime:

class Object(object):
    def __init__(self,*args,**kwargs):
        pass

class StateContainer(Object):
    def __init__(self,parent=None,connector_class=None,*args,**kwargs):
        Object.__init__(self,*args,**kwargs)
        assert [x for x in [parent,connector_class] if x==None]!=[None]
        self.parent=parent
        connector=connector_class(parent,self)

        self.children={} # {child: connector}



class Node(StateContainer):
    def __init__(self,*args,**kwargs):
        StateContainer.__init__(self,*args,**kwargs)
        pass

class TouchedNode(Node):
    def __init__(self,*args,**kwargs):
        Node.__init__(self,*args,**kwargs)
        pass

class UntouchedNode(Node):
    def __init__(self,*args,**kwargs):
        Node.__init__(self,*args,**kwargs)
        pass

class Block(StateContainer):
    def __init__(self,*args,**kwargs):
        StateContainer.__init__(self,*args,**kwargs)
        pass

class Connector(Object):
    def __init__(self,parent,child,*args,**kwargs):
        Object.__init__(self,*args,**kwargs)
        parent.children[child]=self
'''




import pkg_resources

import wx



_icon_bundle = None

def get_icon_bundle():
    
    global _icon_bundle
    
    if _icon_bundle:
        return _icon_bundle

    from . import images as __images_package
    images_package = __images_package.__name__


    icons = []
    for size in [16, 24, 32, 48, 96, 128, 256]:
        file_name = 'icon%s.png' % str(size)
        stream = pkg_resources.resource_stream(
            images_package,
            file_name
        )
        icon = wx.IconFromBitmap(
            wx.BitmapFromImage(
                wx.ImageFromStream(
                    stream,
                    wx.BITMAP_TYPE_PNG
                )
            )
        )
        icons.append(icon)
        
        # todo: should probably be loading the bitmaps from the .ico file, to
        # save on storing all those PNGs.
                    
    
    _icon_bundle = wx.IconBundle()
    
    for icon in icons:
        _icon_bundle.AddIcon(icon)
    
    return _icon_bundle


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the OnlineResourcesMenu class.

See its documentation for more info.
'''


import webbrowser

import wx

from garlicsim_wx.general_misc.cute_menu import CuteMenu


class OnlineResourcesMenu(CuteMenu):
    '''Menu for accessing online resources in the web browser.'''
    def __init__(self, frame):
        super(OnlineResourcesMenu, self).__init__()
        self.frame = frame
        self._build()
    
    def _build(self):
        
        frame = self.frame
        
        
        self.website_button = self.Append(
            -1,
            'Official &website...',
            ' Open the official GarlicSim website in your browser'
        )
        frame.Bind(
            wx.EVT_MENU,
            lambda event: webbrowser.open_new_tab('http://garlicsim.org'),
            self.website_button
        )
        
        
        self.online_documentation_button = self.Append(
            -1,
            'Online &documentation...',
            ''' Open the online version of GarlicSim's documentation in your \
browser'''
        )
        frame.Bind(
            wx.EVT_MENU,
            lambda event: webbrowser.open_new_tab('http://docs.garlicsim.org'),
            self.online_documentation_button
        )
        
        
        self.mailing_lists_button = self.Append(
            -1,
            '&Mailing lists...',
            ''' Open the page with info about GarlicSim mailing lists\
in your browser'''
        )
        frame.Bind(
            wx.EVT_MENU,
            lambda event: webbrowser.open_new_tab(
                'http://garlicsim.org/#mailing_lists'
                ),
            self.mailing_lists_button
        )
        
        
        self.blog_button = self.Append(
            -1,
            '&Blog...',
            ' Open the GarlicSim blog in your browser'
        )
        frame.Bind(
            wx.EVT_MENU,
            lambda event: webbrowser.open_new_tab(
                'http://blog.garlicsim.org'
                ),
            self.blog_button
        )
        

        self.github_button = self.Append(
            -1,
            'Code &repository...',
            ' Open the GitHub code repository for GarlicSim in your browser'
        )
        frame.Bind(
            wx.EVT_MENU,
            lambda event: webbrowser.open_new_tab(
                'http://github.com/cool-RR/GarlicSim'
                ),
            self.github_button
        )
                



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the BitmapViewer class.

See its documentation for more info.
'''

import wx

class BitmapViewer(wx.Panel):
    '''Widget for viewing a bitmap. Similar to StaticBitmap.'''
    def __init__(self, *args, **kwargs):
        wx.Panel.__init__(self, *args, **kwargs)
        #self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        self._bitmap = wx.EmptyBitmap(1, 1)
        self.Bind(wx.EVT_PAINT, self.on_paint)
    
    def on_paint(self, event):
        '''EVT_PAINT handler.'''
        dc = wx.PaintDC(self)
        dc.DrawBitmap(self._bitmap, 0, 0)
        dc.Destroy()
        
    def set_bitmap(self, bitmap):
        '''Set the bitmap that the viewer will display.'''
        self._bitmap = bitmap
        self.Refresh()


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the FileMenu class.

See its documentation for more info.
'''

import wx

from garlicsim_wx.general_misc.cute_menu import CuteMenu

from export_menu import ExportMenu


class FileMenu(CuteMenu):
    '''Menu for file actions: New, open, save...'''
    def __init__(self, frame):
        super(FileMenu, self).__init__()
        self.frame = frame
        self._build()
    
    def _build(self):
        
        frame = self.frame

        
        self.new_button = self.Append(
            wx.ID_NEW, 
            '&New...\tCtrl+N',
            ' Create a new simulation'
        )
        frame.Bind(wx.EVT_MENU, frame.on_new, self.new_button)
        

        self.open_button = self.Append(
            wx.ID_OPEN,
            '&Open...\tCtrl+O',
            ' Open a saved simulation'
        )
        frame.Bind(wx.EVT_MENU, frame.on_open, self.open_button)        
        
        
        # todo: put open recent here

        
        self.AppendSeparator()

        
        self.close_button = self.Append(
            wx.ID_CLOSE,
            '&Close\tCtrl+W',
            ' Close the currently open simulation'
        )
        self.close_button.Enable(False)

        
        self.save_button = self.Append(
            wx.ID_SAVE,
            '&Save\tCtrl+S',
            ' Save the currently open simulation'
        )
        frame.Bind(wx.EVT_MENU, frame.on_save, self.save_button)
        
        
        self.save_as_button = self.Append(
            wx.ID_SAVEAS,
            'Save &as...\tShift+Ctrl+S',
            ' Save the currently open simulation under a different name'
        )
        self.save_as_button.Enable(False)
                
        
        self.AppendSeparator()

        
        self.export_menu = ExportMenu(frame)
        self.export_menu_button = self.AppendMenu(
            -1,
            '&Export',
            self.export_menu,
            ' Export simulation data'
        )
        self.export_menu_button.Enable(False)
        
        
        self.AppendSeparator()
        
        
        self.print_button = self.Append(
            wx.ID_PRINT,
            'Print...\tCtrl+P',
            ' Print the current state of the simulation'
        )
        self.print_button.Enable(False)
        
        
        self.AppendSeparator()
        

        self.exit_button = self.Append(
            wx.ID_EXIT,
            'E&xit',
            ' Close GarlicSim')              
        
        frame.Bind(wx.EVT_MENU, frame.on_exit_menu_button, self.exit_button)


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the App class.

See its documentation for more information.
'''

import functools

import wx

import garlicsim_wx


class App(wx.PySimpleApp):
    '''
    A garlicsim_wx App.
    
    The App is responsible for spawning a Frame.
    '''
    # todo: need to think if i allow frames with no app. on one hand good idea,
    # to allow people to start a garlicsim_wx frame in their own app. on other
    # hand frames will need to know how to start another frame.
    def __init__(self, new_gui_project_simpack_name=None,
                 load_gui_project_file_path=None):
        '''
        Constructor.
        
        In order to start a new simulation on startup, pass the name of the
        simpack as `new_gui_project_simpack_name`.
        
        In order to load a simulation from file on startup, pass the path to the
        file as `load_gui_project_file_path`.
        
        (At most one of these can be done.)
        '''
        self.frame = None
        assert not (new_gui_project_simpack_name and load_gui_project_file_path)
        self.new_gui_project_simpack_name = new_gui_project_simpack_name
        self.load_gui_project_file_path = load_gui_project_file_path
        super(App, self).__init__()
        
    
    def OnInit(self):
        
        frame = garlicsim_wx.Frame(
            parent=None,
            title="GarlicSim",
            size=(1140, 850)
        )
        
        self.frame = frame
        
        self.SetTopWindow(frame)
        
        if self.new_gui_project_simpack_name is not None:
            simpack = __import__(
                self.new_gui_project_simpack_name,
                fromlist=['']
            )
            
            wx.CallAfter(
                functools.partial(
                    self.frame._new_gui_project_from_simpack,
                    simpack
                )
            )
            
        if self.load_gui_project_file_path is not None:
            
            wx.CallAfter(
                functools.partial(
                    self.frame._open_gui_project_from_path,
                    self.load_gui_project_file_path
                )
            )
            
        return True


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the ExportMenu class.

See its documentation for more info.
'''

import wx

from garlicsim_wx.general_misc.cute_menu import CuteMenu

class ExportMenu(CuteMenu):
    '''Menu for exporting simulation data to non-GarlicSim formats.'''
    def __init__(self, frame):
        super(ExportMenu, self).__init__()
        self.frame = frame
        self._build()
    
    def _build(self):
        
        frame = self.frame
        
        
        self.video_button = self.Append(
            -1,
            '&Video',
            ' Export a video sequence showing playback of the simulation'
        )
        self.video_button.Enable(False)
        
        
        self.image_button = self.Append(
            -1,
            '&Image',
            ' Export an image showing a single state in the simulation'
        )
        self.image_button.Enable(False)
                



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''A collection of cursors.'''

# todo: change to use generic caching decorator

import pkg_resources
import wx

from . import images as __images_package
images_package = __images_package.__name__

cached_cursors = {}

def get_open_grab():
    '''Get the "open grab" cursor.'''
    name = 'open_grab'
    file_name = 'open_grab.png'
    hotspot = (8, 8)
    if name in cached_cursors:
        return cached_cursors[name]
    stream = pkg_resources.resource_stream(images_package,
                                           file_name)
    image = wx.ImageFromStream(stream, wx.BITMAP_TYPE_ANY)

    if hotspot is not None:
        image.SetOptionInt(wx.IMAGE_OPTION_CUR_HOTSPOT_X, hotspot[0])
        image.SetOptionInt(wx.IMAGE_OPTION_CUR_HOTSPOT_Y, hotspot[1])
        
    cursor = wx.CursorFromImage(image)
    cached_cursors[name] = cursor
    return cursor


def get_closed_grab():
    '''Get the "closed grab" cursor.'''
    name = 'closed_grab'
    file_name = 'closed_grab.png'
    hotspot = (8, 8)
    if name in cached_cursors:
        return cached_cursors[name]
    stream = pkg_resources.resource_stream(images_package,
                                           file_name)
    image = wx.ImageFromStream(stream, wx.BITMAP_TYPE_ANY)

    if hotspot is not None:
        image.SetOptionInt(wx.IMAGE_OPTION_CUR_HOTSPOT_X, hotspot[0])
        image.SetOptionInt(wx.IMAGE_OPTION_CUR_HOTSPOT_Y, hotspot[1])
        
    cursor = wx.CursorFromImage(image)
    cached_cursors[name] = cursor
    return cursor



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
A package subclassing classes from PyAUI.

This is in order to have an AUI that look and functions the way we like it.
'''

from aui_manager import AuiManager
from aui_dock_art import AuiDockArt
from aui_tab_art import AuiTabArt


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the Frame class.

See its documentation for more information.
'''

from __future__ import with_statement

import os
import sys
import random
import cPickle as pickle_module
import subprocess
import webbrowser
import traceback

import wx
from garlicsim_wx.general_misc.third_party import aui
import pkg_resources

from garlicsim.general_misc import dict_tools
from garlicsim.general_misc import string_tools
from garlicsim_wx.general_misc import thread_timer
from garlicsim_wx.general_misc import wx_tools

import garlicsim
from garlicsim_wx.gui_project import GuiProject
import garlicsim_wx.widgets
import garlicsim_wx.misc
from garlicsim_wx.widgets import workspace_widgets

from . import images as __images_package
images_package = __images_package.__name__


class Frame(wx.Frame):
    '''
    The main window of garlicsim_wx.
    
    This window allows the user to create and manipulate gui projects.
    '''
    def __init__(self, *args, **keywords):
        
        wx.Frame.__init__(self, *args, **keywords)
        
        self.SetDoubleBuffered(True)
        self.SetIcons(garlicsim_wx.misc.icon_bundle.get_icon_bundle())
        
        self.Bind(wx.EVT_CLOSE, self.on_close)        
        
        self.tree_browser = None
        self.seek_bar = None
        self.shell = None
        self.state_repr_viewer = None
        
        self.aui_manager = garlicsim_wx.misc.aui.AuiManager(self)
        '''The aui manager, which manages the workspace widgets.'''
                
        self.gui_project = None
        '''The current gui project.'''
        
        self.CreateStatusBar()
        
        self.__init_menus()
        self.__init_key_handlers()
        
        self.Bind(wx.EVT_CONTEXT_MENU, self.on_context_menu, self)
        
        self.background_timer = thread_timer.ThreadTimer(self)
        
        self.background_timer.start(150)
        
        self.Bind(
            thread_timer.EVT_THREAD_TIMER,
            lambda event: self.sync_crunchers(),
            self.background_timer
        )
        
        self.aui_manager.Update()
        
        self.Show()
        
        self.Maximize()

        
    def __init_menus(self):
        '''Initialize the menu bar and the context menu.'''
        menu_bar = self.menu_bar = garlicsim_wx.misc.MenuBar(self)
        self.SetMenuBar(menu_bar)
        self._recalculate_all_menus()
        self.context_menu = \
            garlicsim_wx.general_misc.cute_menu.CuteMenu.add_menus([
                garlicsim_wx.misc.menu_bar.node_menu.NodeMenu(self),
                garlicsim_wx.misc.menu_bar.block_menu.BlockMenu(self)
            ])
        
        
    def _recalculate_all_menus(self):
        '''Recalculate all the menus, determining in which state they'll be.'''
        try_recalculate = lambda thing: \
            thing._recalculate() if hasattr(thing, '_recalculate') else None
        
        menus_to_recalculate = [menu for (menu, label) in                                 
                                self.menu_bar.GetMenus()]
        
        while menus_to_recalculate:
            menu = menus_to_recalculate.pop()
            for item in menu.GetMenuItems():
                if item.IsSubMenu():
                    menus_to_recalculate.append(item.GetSubMenu())
                else:
                    try_recalculate(item)
            try_recalculate(menu)
        try_recalculate(self.menu_bar)

        
    def __init_key_handlers(self):
        '''Initialize key shortcuts.'''
        
        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        
        def on_home():
            '''Go to root node.'''
            if self.gui_project is not None and \
               self.gui_project.path is not None:
                
                try:
                    self.gui_project.set_active_node(self.gui_project.path[0])
                except LookupError:
                    pass

                
        def on_end():
            '''Go to leaf node.'''
            if self.gui_project is not None and \
               self.gui_project.path is not None:
                
                try:
                    self.gui_project.set_active_node(self.gui_project.path[-1])
                except LookupError:
                    pass

                
        def on_up():
            '''Go one path upwards.'''
            if self.gui_project.path and self.gui_project.active_node:
                try:
                    new_path = self.gui_project.path._get_lower_path(
                        self.gui_project.active_node
                    )
                except LookupError:
                    return
                pseudoclock = self.gui_project.pseudoclock
                self.gui_project.path = new_path
                self.gui_project.path_changed_emitter.emit()
                self.gui_project.set_pseudoclock(pseudoclock)

                
        def on_down():
            '''Go one path downwards.'''
            if self.gui_project.path and self.gui_project.active_node:
                try:
                    new_path = self.gui_project.path._get_higher_path(
                        self.gui_project.active_node
                    )
                except LookupError:
                    return
                pseudoclock = self.gui_project.pseudoclock
                self.gui_project.path = new_path
                self.gui_project.path_changed_emitter.emit()
                self.gui_project.set_pseudoclock(pseudoclock)
                
                
        def on_left():
            '''Go one node backwards.'''
            if self.gui_project is not None and \
               self.gui_project.active_node is not None:
                
                parent = self.gui_project.active_node.parent
                if parent is not None:
                    self.gui_project.set_active_node(parent)
        
                    
        def on_right():
            '''Go one node forward.'''
            if self.gui_project is not None and \
               self.gui_project.path is not None and \
               self.gui_project.active_node is not None:
                
                try:
                    child = self.gui_project.path.next_node\
                          (self.gui_project.active_node)
                    self.gui_project.set_active_node(child)
                except LookupError:
                    pass
            
                
        def on_command_left():
            '''Go five nodes backwards.'''
            if self.gui_project is not None and \
               self.gui_project.active_node is not None:
                
                node = self.gui_project.active_node.get_ancestor(5, round=True)
                if node:
                    self.gui_project.set_active_node(node)
        
                    
        def on_command_right():
            '''Go five nodes forward.'''
            if self.gui_project is not None and \
               self.gui_project.path is not None and \
               self.gui_project.active_node is not None:
                
                current = self.gui_project.active_node
                for i in xrange(5):
                    try:
                        current = self.gui_project.path.next_node(current)
                    except LookupError:
                        pass
                self.gui_project.set_active_node(current)
        
                
        def on_page_up():
            '''Go 20 nodes backwards.'''
            if self.gui_project is not None and \
               self.gui_project.active_node is not None:
                
                node = self.gui_project.active_node.get_ancestor(20, round=True)
                if node:
                    self.gui_project.set_active_node(node)
        
                    
        def on_page_down():
            '''Go 20 nodes forward.'''
            if self.gui_project is not None and \
               self.gui_project.path is not None and \
               self.gui_project.active_node is not None:
                
                current = self.gui_project.active_node
                for i in xrange(20):
                    try:
                        current = self.gui_project.path.next_node(current)
                    except LookupError:
                        pass
                self.gui_project.set_active_node(current)

                
        def on_space():
            '''Toggle onscreen playback.'''
            if self.gui_project:
                self.gui_project.toggle_playing()
        
                
        def on_return():
            '''Finalize the active node, if it's in editing.'''
            if self.gui_project and self.gui_project.active_node.still_in_editing:
                self.gui_project.finalize_active_node()

        
                
        
        self.key_handlers = {
            wx_tools.Key(wx.WXK_HOME): on_home,
            wx_tools.Key(wx.WXK_END): on_end,
            wx_tools.Key(wx.WXK_UP): on_up,
            wx_tools.Key(wx.WXK_DOWN): on_down,
            wx_tools.Key(wx.WXK_LEFT): on_left,
            wx_tools.Key(wx.WXK_RIGHT): on_right,
            wx_tools.Key(wx.WXK_LEFT, cmd=True): on_command_left,
            wx_tools.Key(wx.WXK_RIGHT, cmd=True): on_command_right,
            wx_tools.Key(wx.WXK_PAGEUP): on_page_up,
            wx_tools.Key(wx.WXK_PAGEDOWN): on_page_down,
            wx_tools.Key(wx.WXK_SPACE): on_space,
            wx_tools.Key(wx.WXK_RETURN): on_return,
        }    
            
        
    def on_close(self, event):
        '''Close the frame.'''
        if self.gui_project:
            self.gui_project.stop_playing()
        self.aui_manager.UnInit()
        self.Destroy()
        garlicsim_wx.general_misc.cute_base_timer.CuteBaseTimer.\
                    stop_timers_by_frame(self)
        event.Skip()        
        self.background_timer.stop()

        
    def finalize_active_node(self, e=None):
        '''Finalize editing of the active node in the active gui project.'''
        assert self.gui_project
        return self.gui_project.finalize_active_node()

    
    def on_new(self, event=None):
        '''Create a new gui project.'''        
        
        if self.gui_project is not None:
            
            if hasattr(sys, 'frozen'):
                program_to_run = [sys.executable]
                we_are_main_program = 'GarlicSim' in sys.executable
            else:
                main_script = os.path.abspath(sys.argv[0])
                program_to_run = [sys.executable, main_script]
                we_are_main_program = ('run_gui' in main_script) or \
                                    ('garlicsim_wx' in main_script)
            
            if not we_are_main_program:
                warning_dialog = \
                    garlicsim_wx.widgets.misc.NotMainProgramWarningDialog(self)
                if warning_dialog.ShowModal() != wx.ID_YES:
                    return
        
        dialog = garlicsim_wx.widgets.misc.SimpackSelectionDialog(self)
        
        if dialog.ShowModal() == wx.ID_OK:
            simpack = dialog.get_simpack_selection()
        else:
            dialog.Destroy()
            return
        dialog.Destroy()

        
        if self.gui_project is None:
            self._new_gui_project_from_simpack(simpack)
        else:    
                
            program_to_run.append('__garlicsim_wx_new=%s' % simpack.__name__)
         
            subprocess.Popen(program_to_run)
            
            return
            
    def _new_gui_project_from_simpack(self, simpack):
        '''
        Start a new gui project, given the simpack to start it with.
        
        Internal use.
        '''
        assert self.gui_project is None
        gui_project = GuiProject(simpack, self)
        self.__setup_gui_project(gui_project)

        
    def on_exit_menu_button(self, event):
        '''Exit menu button handler.'''
        self._post_close_event()

        
    def _post_close_event(self):
        '''Post a close event to the frame.'''
        event = wx.PyEvent(self.Id)
        event.SetEventType(wx.wxEVT_CLOSE_WINDOW)
        wx.PostEvent(self, event)
        
        
    def sync_crunchers(self):
        '''
        Take work from the crunchers, and give them new instructions if needed.
                
        (This is a wrapper that calls the sync_crunchers method of all the
        gui projects.)
        
        Talks with all the crunchers, takes work from them for implementing
        into the tree, retiring crunchers or recruiting new crunchers as
        necessary.
        
        Returns the total amount of nodes that were added to each gui project's
        tree.
        '''
        nodes_added = self.gui_project.sync_crunchers() \
                    if self.gui_project else 0
        
        if nodes_added > 0:
            pass#self.Refresh()
        
        return nodes_added
    
    def __setup_gui_project(self, gui_project):
        '''
        Setup a newly-created gui project.
        
        Internal use.
        '''
        
        self.gui_project = gui_project
        
        # todo: should create StateReprViewer only if the simpack got no
        # workspace widgets
        
        self.tree_browser = workspace_widgets.TreeBrowser(self)
        self.aui_manager.AddPane(
            self.tree_browser,
            aui.AuiPaneInfo()\
            .Bottom().Row(0)\
            .BestSize(1000, 100).MinSize(200, 50).MaxSize(10000, 250)\
            .Caption(self.tree_browser.get_uppercase_name())
            .Floatable(False)\
            .CloseButton(False)
        )
        
        self.playback_controls = workspace_widgets.PlaybackControls(self)
        self.aui_manager.AddPane(
            self.playback_controls,
            aui.AuiPaneInfo()\
            .Bottom()\
            .BestSize(184, 128).MinSize(184, 128).MaxSize(184, 128)\
            .Caption(self.playback_controls.get_uppercase_name())
            .Resizable(False)\
            .CloseButton(False)        
        )
        
        self.seek_bar = workspace_widgets.SeekBar(self)
        self.aui_manager.AddPane(
            self.seek_bar,
            aui.AuiPaneInfo()\
            .Bottom().Row(1)\
            .BestSize(600, 40).MinSize(200, 40).MaxSize(10000, 100)\
            .Caption(self.seek_bar.get_uppercase_name())
            .Floatable(False)\
            .CloseButton(False)
        )
        
        self.shell = workspace_widgets.Shell(self)
        self.aui_manager.AddPane(
            self.shell,
            aui.AuiPaneInfo()\
            .Right().Row(0)\
            .BestSize(400, 600)\
            .Caption(self.shell.get_uppercase_name())
            .MaximizeButton(True)\
            .CloseButton(False)
        )
        
        """
        self.state_repr_viewer = workspace_widgets.StateReprViewer(self)
        self.aui_manager.AddPane(
            self.state_repr_viewer,
            aui.AuiPaneInfo()\
            .BestSize(300, 300)\
            .MaximizeButton(True)\
            .Center()\
            .Caption(self.state_repr_viewer.get_uppercase_name())
            .Floatable(False)\
            .CloseButton(False)
        )
        """
        settings_wx = self.gui_project.simpack_wx_grokker.settings
        

        big_widget_class = settings_wx.BIG_WORKSPACE_WIDGETS[0] if \
                         settings_wx.BIG_WORKSPACE_WIDGETS else \
                         workspace_widgets.StateReprViewer

        self.big_widget = big_widget_class(self)
        self.aui_manager.AddPane(
            self.big_widget,
            aui.AuiPaneInfo()\
            .BestSize(300, 300)\
            .MaximizeButton(True)\
            .Center()\
            .Caption(self.big_widget.get_uppercase_name())
            .Floatable(False)\
            .CloseButton(False)
        )
        
        if isinstance(self.big_widget, workspace_widgets.StateReprViewer):
            self.state_repr_viewer = self.big_widget
        
        """
        big_widget_classes = \
            settings_wx.BIG_WORKSPACE_WIDGETS #+ \
        #    [workspace_widgets['StateReprViewer']]
        
        self.big_widgets = []
        # todo: not the right way, should be easy listing of all widget
        
        
        for i, BigWidget in enumerate(big_widget_classes):
            big_widget = BigWidget(self)
            self.aui_manager.AddPane(
                big_widget,
                aui.AuiPaneInfo()\
                .BestSize(300, 300)\
                .MaximizeButton(True)\
                .Center()\
                .Caption(big_widget.get_uppercase_name())\
                .Floatable(False)\
                .CloseButton(False),
                target=self.state_repr_viewer.get_aui_pane_info()
            )
            #.NotebookPage(notebook_id, i)\
            self.big_widgets.append(big_widget)

        """
        
        self.aui_manager.Update()
        
        self.gui_project.emitter_system.top_emitter.emit()
        
    
    def on_open(self, event=None):
        '''Raise a dialog for opening a gui project from file.'''
        
        if self.gui_project is not None:
            
            if hasattr(sys, 'frozen'):
                program_to_run = [sys.executable]
                we_are_main_program = 'GarlicSim' in sys.executable
            else:
                main_script = os.path.abspath(sys.argv[0])
                program_to_run = [sys.executable, main_script]
                we_are_main_program = ('run_gui' in main_script) or \
                                    ('garlicsim_wx' in main_script)
            
            if not we_are_main_program:
                dialog = \
                    garlicsim_wx.widgets.misc.NotMainProgramWarningDialog(self)
                if dialog.ShowModal() != wx.ID_YES:
                    return
                
        wildcard = 'GarlicSim Simulation Pickle (*.gssp)|*.gssp|All files (*)|*|'
        
        # Todo: something more sensible here. Ideally should be last place you
        # saved in, but for starters can be desktop.
        folder = os.getcwd()
        
        gui_project_vars = None

        open_dialog = wx.FileDialog(self, message='Choose a file',
                                    defaultDir=folder, defaultFile='',
                                    wildcard=wildcard, style=wx.OPEN)
        if open_dialog.ShowModal() == wx.ID_OK:
            path = open_dialog.GetPath()
            
            if self.gui_project is None:
                self._open_gui_project_from_path(path)
            else:
                if hasattr(sys, 'frozen'):
                    program = [sys.executable]
                else:
                    program = [sys.executable, os.path.abspath(sys.argv[0])]
                    # Todo: what if some other program is launching my code?
                    
                program.append(path)
             
                subprocess.Popen(program)
                        
        
    
    def _open_gui_project_from_path(self, path):
        '''
        Open a gui project saved to a file specified by `path`.
        
        Internal use.
        '''
        
        old_recursion_limit = sys.getrecursionlimit()
            
        try:
            sys.setrecursionlimit(10000)
            with open(path, 'rb') as my_file:
                gui_project_vars = pickle_module.load(my_file)
                
        except Exception, exception:
            dialog = wx.MessageDialog(
                self,
                'Error opening file:\n' + traceback.format_exc(),
                style=(wx.OK | wx.ICON_ERROR)
            )
            dialog.ShowModal()
            return
        
        finally:
            sys.setrecursionlimit(old_recursion_limit)
                
        if gui_project_vars:
            try:
                gui_project = GuiProject.load_from_vars(self, gui_project_vars)
            except Exception, exception:
                error_dialog = wx.MessageDialog(
                    self,
                    'Error opening file:\n' + traceback.format_exc(),
                    style=(wx.OK | wx.ICON_ERROR)
                )
                error_dialog.ShowModal()
                error_dialog.Destroy()
                
            self.__setup_gui_project(gui_project)

    
    
    def on_save(self, event=None):
        '''Raise a dialog for saving a gui project to file.'''
        
        assert self.gui_project is not None
        wildcard = 'GarlicSim Simulation Pickle (*.gssp)|*.gssp|All files (*)|*|'
        folder = os.getcwd()
        
        save_dialog = wx.FileDialog(self, message='Save file as...',
                                 defaultDir=folder, defaultFile='',
                                 wildcard=wildcard,
                                 style=wx.SAVE | wx.OVERWRITE_PROMPT)
        if save_dialog.ShowModal() == wx.ID_OK:
            path = save_dialog.GetPath()
            
            old_recursion_limit = sys.getrecursionlimit()
            
            try:
                sys.setrecursionlimit(10000)
                with open(path, 'wb') as my_file:
                    picklable_vars = self.gui_project.__getstate__()
                    pickle_module.dump(picklable_vars, my_file, protocol=2)

            except Exception, exception:
                error_dialog = wx.MessageDialog(
                    self,
                    'Error saving to file:\n' + traceback.format_exc(),
                    style=(wx.OK | wx.ICON_ERROR)
                )
                error_dialog.ShowModal()
                error_dialog.Destroy()
            
            finally:
                sys.setrecursionlimit(old_recursion_limit)
            
            
        save_dialog.Destroy()
    
    """    
    def delete_gui_project(self,gui_project):
        I did this wrong.
        self.gui_projects.remove(gui_project)
        self.notebook.AddPage(gui_project.main_window,"zort!")
        self.notebook.DeletePage(0)
        del gui_project
    """
    
    
    def on_key_down(self, event):
        '''wx.EVT_KEY_DOWN handler.'''
        key = wx_tools.Key.get_from_key_event(event)
        handler = self.key_handlers.get(key, None)
        if handler:
            handler()
        else:
            event.Skip()
            
            
    def on_context_menu(self, event):
        '''wx.EVT_CONTEXT_MENU handler.'''
        abs_position = event.GetPosition()
        if abs_position == wx.DefaultPosition:
            position = (0, 0)
        else:
            position = self.ScreenToClient(abs_position)
            
        self.PopupMenu(self.context_menu, position)
        
    
    



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the CuteBaseTimer class.

See its documentation for more information.
'''

class CuteBaseTimer(object):
    '''A base class for timers, allowing easy central stopping.'''    
    __timers = [] # todo: change to weakref list
    
    def __init__(self, parent):
        self.__parent = parent
        CuteBaseTimer.__timers.append(self)
        
        
    @staticmethod # should be classmethod?
    def stop_timers_by_frame(frame):
        '''Stop all the timers that are associated with the given frame.'''
        for timer in CuteBaseTimer.__timers:
            ancestor = timer.__parent
            while ancestor is not None:
                if ancestor == frame:
                    timer.Stop()
                    break
                ancestor = ancestor.GetParent()
    


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This module defines the SimpackWxGrokker and the Settings classes.

See their documentation for more info.
'''

from .settings import Settings
from .simpack_wx_grokker import SimpackWxGrokker


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the CreateMenu class.

See its documentation for more info.
'''

import wx

from garlicsim_wx.general_misc.cute_menu import CuteMenu


class CreateMenu(CuteMenu):
    '''Menu for creating new objects.'''
    def __init__(self, frame):
        super(CreateMenu, self).__init__()
        self.frame = frame
        self._build()
    
    def _build(self):
        
        frame = self.frame
        
        self.create_state_button = self.Append(
            -1, 
            'Create &state...',
            ' Create a new state, which will become a root node in the tree'
        )
        frame.Bind(
            wx.EVT_MENU,
            lambda event: frame.gui_project.make_state_creation_dialog(),
            self.create_state_button
        )
        
    
        self.AppendSeparator()
        
        
        self.create_step_profile_button = self.Append(
            -1, 
            'Create step &profile...',
            ''' Create a new step profile, which can modify the world rules \
under which the simulation crunches'''
        )
        self.create_step_profile_button.Enable(False)
        
        


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''A collection of cursors.'''

from cursor_collection import *


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines an EmitterSystem, which offers some benefits over Emitter.

See documentation of EmitterSystem for more info.
'''


from emitter_system import EmitterSystem
from emitter import Emitter


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This module defines the Settings class.

See its documentation for more info.
'''

import garlicsim_wx

class Settings(object):
    '''A set of settings for a simpack_wx.'''
    #todo: subclass from a pretty vars-shower
    def __init__(self):
        
        self.BIG_WORKSPACE_WIDGETS = []
        '''Widgets to show in the middle of the frame.'''
        
        self.SMALL_WORKSPACE_WIDGETS = []
        '''Small widgets to show.'''
        
        self.SEEK_BAR_GRAPHS = []
        '''
        List of scalar functions that should be shown as graphs in the seek bar.
        
        These may be either scalar state functions or scalar history functions.
        '''
        
        self.STATE_CREATION_DIALOG = \
            garlicsim_wx.widgets.misc.StateCreationDialog
        '''Dialog for creating a root state.'''



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the SnapMap class.

See its documentation for more info.
'''

from __future__ import division
from garlicsim.general_misc import misc_tools


FUZZ = 0.001
'''
The fuzziness of floating point numbers.

If two floats have a distance of less than FUZZ, we may treat them as identical.
'''


class SnapMap(object):
    '''
    Map for deciding which angle the knob will have when mouse-dragging.
    
    
    Here we have three "scales" we are playing in:
    
    1. The "ratio" scale. See documenation on Knob for that one. This controls
       the angle of the knob and the actual value of the final variable.
    
    2. The "y" scale. This is the `y` reading of the mouse on the screen.
    
    3. The "pos" scale. This is a convenient mediator between the first two. It
       is reversed from "y", because on the screen a higher number of y means
       "down", and that's just wrong. Also, it has some translation.
       
    '''
    def __init__(self, snap_point_ratios, base_drag_radius,
                 snap_point_drag_well, initial_y, initial_ratio):
        
        assert snap_point_ratios == sorted(snap_point_ratios)
        
        self.snap_point_ratios = snap_point_ratios
        '''Ordered list of snap points, as ratios.'''
        
        self.base_drag_radius = base_drag_radius
        '''
        The base drag radius, in pixels.
        
        This number is the basis for calculating the height of the area in which
        the user can play with the mouse to turn the knob. Beyond that area the
        knob will be turned all the way to one side, and any movement farther
        will have no effect.
        
        If there are no snap points, the total height of that area will be `2 *
        self.base_drag_radius`.
        '''
        
        self.snap_point_drag_well = snap_point_drag_well
        '''
        The height of a snap point's drag well, in pixels.
        
        This is the height of the area on the screen in which, when the user
        drags to it, the knob will have the value of the snap point.
        
        The bigger this is, the harder the snap point "traps" the mouse.
        '''
        
        self.initial_y = initial_y
        '''The y that was recorded when the user started dragging.'''
        
        self.initial_ratio = initial_ratio
        '''The ratio that was recorded when the user started dragging.'''
        
        self.initial_pos = self.ratio_to_pos(initial_ratio)
        '''The pos that was recorded when the user started dragging.'''
        
        self.max_pos = base_drag_radius * 2 + \
            len(snap_point_ratios) * snap_point_drag_well
        '''The maximum that a pos number can reach before it gets truncated.'''
        
        self._make_snap_point_pos_starts()
            
    
    ###########################################################################
    # # # # Converters:
    ############
    
    def ratio_to_pos(self, ratio):
        '''Convert from ratio to pos.'''
        assert (- 1 - FUZZ) <= ratio <= 1 + FUZZ
        n_snap_points_from_bottom = self._get_n_snap_points_from_bottom(ratio)
        padding = n_snap_points_from_bottom * self.snap_point_drag_well
        distance_from_bottom = ratio - (-1)
        result = padding + distance_from_bottom * self.base_drag_radius
        return result

    def pos_to_y(self, pos):
        '''Convert from pos to y.'''
        assert 0 - FUZZ <= pos <= self.max_pos + FUZZ
        relative_pos = (pos - self.initial_pos)
        return self.initial_y - relative_pos
        # doing minus because y is upside down
    
    def y_to_pos(self, y):
        '''Convert from y to pos.'''
        relative_y = (y - self.initial_y)

        # doing minus because y is upside down
        pos = self.initial_pos - relative_y
        
        if pos < 0:
            pos = 0
        if pos > self.max_pos:
            pos = self.max_pos
        
        return pos
        
        
    def pos_to_ratio(self, pos):
        '''Convert from pos to ratio.'''
        assert 0 - FUZZ <= pos <= self.max_pos + FUZZ
        
        snap_point_pos_starts_from_bottom = [
            p for p in self.snap_point_pos_starts if p <= pos
        ]
        
        padding = 0
        
        if snap_point_pos_starts_from_bottom:

            candidate_for_current_snap_point = \
                snap_point_pos_starts_from_bottom[-1]
        
            distance_from_candidate = (pos - candidate_for_current_snap_point)
            
            if distance_from_candidate < self.snap_point_drag_well:
                
                # It IS the current snap point!
                
                snap_point_pos_starts_from_bottom.remove(
                    candidate_for_current_snap_point
                )
                
                padding += distance_from_candidate
        
        padding += \
            len(snap_point_pos_starts_from_bottom) * self.snap_point_drag_well
        
        
        ratio = ((pos - padding) / self.base_drag_radius) - 1
        
        assert (- 1 - FUZZ) <= ratio <= 1 + FUZZ
        
        return ratio
        
    
    def ratio_to_y(self, ratio):
        '''Convert from ratio to y.'''
        return self.pos_to_y(self.ratio_to_pos(ratio))
    
    def y_to_ratio(self, y):
        '''Convert from y to ratio.'''
        return self.pos_to_ratio(self.y_to_pos(y))
    
    ###########################################################################
    
    def _get_n_snap_points_from_bottom(self, ratio):
        '''Get the number of snap points whose ratio is lower than `ratio`.'''
        raw_list = [s for s in self.snap_point_ratios
                    if -1 <= s <= (ratio + FUZZ)]
        
        if not raw_list:            
            return 0
        else: # len(raw_list) >= 1
            counter = 0
            counter += len(raw_list[:-1])
            last_snap_point = raw_list[-1]
            ratio_in_last_snap_point = (abs(last_snap_point - ratio) < FUZZ)
            if ratio_in_last_snap_point:
                counter += 0.5
            else:
                counter += 1
            return counter    
    
    
    def _make_snap_point_pos_starts(self):
        '''
        Make a list with a "pos start" for each snap point.
        
        A "pos start" is the lowest point, in pos scale, of a snap point's drag
        well.

        The list is not returned, but is stored as the attribute
        `.snap_point_pos_starts`.
        '''
        
        self.snap_point_pos_starts = []
        
        for i, ratio in enumerate(self.snap_point_ratios):
            self.snap_point_pos_starts.append(
                (1 + ratio) * self.base_drag_radius + \
                i * self.snap_point_drag_well
            )
            
    
    """
    def _debug_ratio_to_pos(self, step=0.1):
        if step is None: step = 10
        
        return [(i, self.ratio_to_pos(i)) for i in
                misc_tools.frange(-1, 1, step)]
    """
        

"""
if __name__ == '__main__':
    s=SnapMap([-0.3,0,0.5,0.7],100,100,0,0)
"""


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
A little module for cataloging strings.
'''

stringsaver_catalog = []

def s2i(string):
    '''
    If the string isn't cataloged already, catalog it.
    
    In any case, returns the number associated with the string.
    '''
    global stringsaver_catalog
    if string in stringsaver_catalog:
        return stringsaver_catalog.index(string) + 1
    else:
        stringsaver_catalog.append(string)
        return stringsaver_catalog.index(string) + 1


def i2s(integer):
    '''
    Return the string cataloged under the given integer.
    '''
    return stringsaver_catalog[integer - 1]



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the BlockMenu class.

See its documentation for more info.
'''

import wx

from garlicsim_wx.general_misc.cute_menu import CuteMenu

class BlockMenu(CuteMenu):
    '''A menu for manipulating the active block.'''
    def __init__(self, frame):
        super(BlockMenu, self).__init__()
        self.frame = frame
        self._build()
    
    def _build(self):
        
        frame = self.frame
        
       
        self.split_button = self.Append(
            -1,
            '&Split active block...',
            " Split the active block into two separate blocks"
        )
        self.split_button.Enable(False)
        
        
        self.scatter_button = self.Append( # todo: rename
            -1,
            'S&catter active block...',
            ' Scatter the active block, leaving all its nodes blockless'
        )
        self.scatter_button.Enable(False)
        
        


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Bootstrap module for garlicsim_wx.

It checks all prerequisites are installed.
'''

import warnings
import sys


if sys.version_info[0] >= 3:
    raise Exception('''This package is not compatible with Python 3.x.''')
if sys.version_info[1] <= 4:
    raise Exception('''This package requires Python 2.5 and upwards. (Not \
including 3.x).''')


def __check_prerequisites():
    '''
    Check that all modules required for garlicsim_wx are installed.
    
    Returns a list of some imported modules: A reference to this list should be
    kept alive so to prevent the imported modules from being garbage-collected,
    which would cause Python to load them twice, which would needlessly increase
    startup time.
    '''
    
    modules = []
    
    class MissingModule(Exception):
        '''A required module is not found.'''
    
    def check_garlicsim():
        try:
            import garlicsim
            return [garlicsim]
        except ImportError:
            raise MissingModule('''garlicsim is \
required, but it's not currently installed on your system. Go to \
http://garlicsim.org and follow the instructions for installation.''')
        
    
    def check_wx():
        try:
            import wx
            if wx.__version__ < '2.8.10.1':
                warnings.warn('''You have wxPython version %s installed, \
while version 2.8.10.1 is needed. This program will try to run now, and if \
you're lucky it will work alright, but if any problem comes up, try upgrading \
wxPython. To do that, download and install the latest version from \
http://wxpython.org''' % wx.__version__)
            return [wx]
        except ImportError:
            raise MissingModule('''wxPython (version 2.8.10.1 and upwards) is \
required, but it's not currently installed on your system. Please find it \
online and install it, then try again.''')
    

    checkers = [check_garlicsim, check_wx]
    
    for checker in checkers:
        modules += checker()
    
    return modules

__modules_list = __check_prerequisites()



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This module defines the SimpackWxGrokker class.

See its documentation for more info.
'''

import types

from garlicsim.general_misc import import_tools
import garlicsim.general_misc.caching

import garlicsim_wx

from .settings import Settings


class SimpackWxGrokker(object):
    '''Encapsulates a simpack_wx and gives useful information and tools.'''
    
    __metaclass__ = garlicsim.general_misc.caching.CachedType
    
    def __init__(self, simpack):
        self.simpack = simpack
        
        if isinstance(simpack, types.ModuleType):
            simpack_wx_module_name = ''.join((
                self.simpack.__name__,
                '.wx'
            ))
        
        import_tools.import_if_exists(simpack_wx_module_name)
        # This imports the `wx` submodule, if it exists, but it does *not* keep
        # a reference to it. We'll access `wx` as an attribute of the simpack
        # below.
            
        try:
            self.simpack_wx = self.simpack.wx
        except AttributeError:
            self.simpack_wx = None
            #raise Exception('''Simpack has no wx''') # todo: edit this
        
        self.__init_analysis_settings()
    
        
    
    def __init_analysis_settings(self):
        '''Analyze the simpack_wx to produce a Settings object.'''
        # todo: consider doing this in Settings.__init__
        
        # We want to access the `.settings` of our simpack_wx, but we don't know if
        # our simpack_wx is a module or some other kind of object. So if it's a
        # module, we'll try to import `settings`.
        
        self.settings = Settings()        
        
        if isinstance(self.simpack_wx, types.ModuleType) and \
           not hasattr(self.simpack_wx, 'settings'):
            
            # The `if` that we did here means: "If there's reason to suspect
            # that self.simpack_wx.settings is a module that exists but hasn't
            # been imported yet."
                
            settings_module_name = ''.join((
                self.simpack_wx.__name__,
                '.settings'
            ))
            
            import_tools.import_if_exists(settings_module_name)
            # This imports the `settings` submodule, if it exists, but it does
            # *not* keep a reference to it. We'll access `settings` as an
            # attribute of the simpack_wx below.
            
        # Checking if there are original settings at all. If there aren't, we're
        # done.
        if self.simpack_wx and hasattr(self.simpack_wx, 'settings'):
            
            original_settings = getattr(self.simpack_wx, 'settings')
        
            for (key, value) in vars(self.settings).iteritems():
                if hasattr(original_settings, key):
                    actual_value = getattr(original_settings, key)
                    setattr(self.settings, key, actual_value)
            # todo: currently throws away unrecognized attributes from the
            # simpack's settings.
                
    


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the MenuBar class.

See its documentation for more info.
'''

import wx

from file_menu import FileMenu
from edit_menu import EditMenu
from create_menu import CreateMenu
from node_menu import NodeMenu
from block_menu import BlockMenu
from window_menu import WindowMenu
from help_menu import HelpMenu


class MenuBar(wx.MenuBar):
    '''The main menubar of garlicsim_wx.'''
    def __init__(self, frame):
        super(MenuBar, self).__init__()
        self.frame = frame
        
        is_mac = (wx.Platform == '__WXMAC__')
        
        self.file_menu = FileMenu(frame)
        self.Append(self.file_menu, '&File')
        
        self.edit_menu = EditMenu(frame)
        self.Append(self.edit_menu, '&Edit')
        # This disables a menu from the bar:
        # self.EnableTop(self.FindMenu('Edit'), False)
        # Logically it makes sense, but it makes it hard to see all the options
        # in the menu, so at least for now I'm not doing it.
        
        self.create_menu = CreateMenu(frame)
        self.Append(self.create_menu, '&Create')
        self.create_menu._recalculate = lambda: self.EnableTop(
            self.FindMenu('Create'), 
            frame.gui_project is not None
        )
        
        self.node_menu = NodeMenu(frame)
        self.Append(self.node_menu, '&Node')
        self.node_menu._recalculate = lambda: self.EnableTop(
            self.FindMenu('Node'), 
            frame.gui_project is not None and \
            frame.gui_project.active_node is not None
        )
        
        self.block_menu = BlockMenu(frame)
        self.Append(self.block_menu, '&Block')
        self.block_menu._recalculate = lambda: self.EnableTop(
            self.FindMenu('Block'), 
            frame.gui_project is not None and \
            frame.gui_project.active_node is not None and \
            frame.gui_project.active_node.block is not None
        )
        
        self.window_menu = WindowMenu(frame)
        title_of_window_menu = '&Workspace' if is_mac else '&Window'
        self.Append(self.window_menu, title_of_window_menu)
        
        self.help_menu = HelpMenu(frame)
        title_of_help_menu = 'GarlicSim &Help' if is_mac else '&Help'
        self.Append(self.help_menu, title_of_help_menu)


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the NodeMenu class.

See its documentation for more info.
'''

import wx

from garlicsim_wx.general_misc.cute_menu import CuteMenu


class NodeMenu(CuteMenu):
    '''Menu for manipulating the active node.'''
    def __init__(self, frame):
        super(NodeMenu, self).__init__()
        self.frame = frame
        self._build()
    
    def _build(self):
        
        frame = self.frame
       
        self.fork_by_editing_button = self.Append(
            -1,
            'Fork by &editing',
            ''' Fork the simulation by making a copy of the active node and \
editing it'''
        )
        
        frame.Bind(
            wx.EVT_MENU,
            lambda event: frame.gui_project.fork_by_editing(),
            self.fork_by_editing_button
        )

        
        self.fork_by_crunching_button = self.Append(
            -1,
            'Fork by &crunching',
            ' Fork the simulation by crunching from the active node'
        )
        frame.Bind(
            wx.EVT_MENU,
            lambda event: frame.gui_project.fork_by_crunching(),
            self.fork_by_crunching_button
        )

        
        self.AppendSeparator()
        
        
        self.properties_button = self.Append(
            -1,
            'Node &properties...',
            " See the active node's properties"
        )
        self.properties_button.Enable(False)        
        
        
        self.AppendSeparator()
        
        
        self.delete_button = self.Append(
            -1,
            '&Delete active node...',
            ' Delete the active node'
        )
        self.delete_button.Enable(False)
                



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This module defines the get_lines function. See its documentation for more
info.
'''

def get_lines(start, end, thing):
    '''
    Get a list of all "round" numbers between start and end.
    
    with thing=0, round numbers = 1, 2, 3...
    with thing=1, round numbers = 10, 20, 30...
    with thing=-1, round numbers = 0.1, 0.2, 0.3...
    etc.
    '''
    mything = int(round(thing))
    mystart = round(start,mything)
    if mystart < start:
        mystart += 10 ** mything
    '''
    myend=round(end,-thing)
    if myend>end:
        myend-=10**thing
    '''

    result = []
    current = mystart
    while current < end:
        result.append(round(current, -mything))
        current += 10 ** mything
    return result



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This package defines generic wxPython widgets.
'''

from knob import Knob
from cute_dialog import CuteDialog


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This module defines the AboutDialog class.

See its documentation for more info.
'''

import time
import webbrowser

import pkg_resources
import wx.html

from garlicsim_wx.general_misc import wx_tools
import garlicsim_wx.general_misc.cute_timer

import garlicsim_wx
from bitmap_viewer import BitmapViewer

from . import images as __images_package
images_package = __images_package.__name__


class AboutDialog(wx.Dialog):
    '''An About dialog for GarlicSim.'''
    def __init__(self, frame):
   
        wx.Dialog.__init__(self, frame, title='About GarlicSim',
                           size=(628, 600))
        
        self.SetBackgroundColour(wx.Color(212, 208, 200))
        
        self.SetDoubleBuffered(True)
        
        self.frame = frame

        v_sizer = wx.BoxSizer(wx.VERTICAL)
        

        self._original_image = wx.ImageFromStream(
            pkg_resources.resource_stream(
                images_package,
                'about.png'
            )
        )
        
        self.bitmap_viewer = BitmapViewer(self, size=(627, 271))
        v_sizer.Add(self.bitmap_viewer, 0)
        
        #self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        #self.SetBackgroundColour(wx_tools.get_background_color())
        
        
        self.html_window = wx.html.HtmlWindow(self, size=(628, 270))
        v_sizer.Add(self.html_window, 0)
        
        self.html_window.SetPage(
            '''
            <html>
                <body bgcolor="#d4d0c8">
                    <div align="center"> <font size="1">
                        &copy; 2009-2010 Ram Rachum (a.k.a. cool-RR)
                        <br />                        
                        No part of this program may be used, copied or
                        distributed without explicit written permission from Ram
                        Rachum.
                        <br />
                    </font></div>
                    <div> 
                        GarlicSim is a platform for writing, running and
                        analyzing computer simulations. It is general enough to
                        handle any kind of simulation: Physics, game theory,
                        epidemic spread, electronics, etc.<br />
                        <font size="1"><br /></font>
                        <b>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Version %s</b>
                        &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Website:
                        <a href="http://garlicsim.org">http://garlicsim.org</a>
                    </div>
                    <div>
                        <font size="1"><br /></font>
                        I would like to thank the open source community for
                        making this program possible. This includes the
                        developers of Python, Psyco, wxPython, wxWidgets,
                        Mayavi, git, and so many others... And more thanks to
                        the many people who spent many hours helping me out
                        with various issues, on mailing lists such as
                        wxpython-users and on the StackOverflow website.
                    </div>
                </body>
            </html>
            ''' % garlicsim_wx.__version__
        )
        
        self.html_window.Bind(
            wx.html.EVT_HTML_LINK_CLICKED, 
            lambda event: webbrowser.open_new_tab(
                event.GetLinkInfo().GetHref()
                ),
            self.html_window
        )

        
        self.button_sizer = button_sizer = wx.StdDialogButtonSizer()
        self.ok_button = wx.Button(self, wx.ID_OK,
                                   "Let's get back to simulating!")
        self.ok_button.SetDefault()
        button_sizer.SetAffirmativeButton(self.ok_button)
        self.Bind(wx.EVT_BUTTON, self.on_ok, self.ok_button)
        button_sizer.AddButton(self.ok_button)
        button_sizer.Realize()
        button_sizer.SetMinSize((500, -1))
        v_sizer.Add(button_sizer, 0)
        
        
        self.SetSizer(v_sizer)
        self.Layout()

        
        self.timer = garlicsim_wx.general_misc.cute_timer.CuteTimer(self)
        self.timer.Start(40)
        self.Bind(wx.EVT_TIMER, self.on_timer, self.timer)
        
        self._rotate_image_hue()

        
    def on_ok(self, e=None):
        '''Do 'Okay' on the dialog.'''

        self.EndModal(wx.ID_OK)

        
    def on_timer(self, event):
        '''wx.EVT_TIMER handler.'''
        self._rotate_image_hue()

        
    def _rotate_image_hue(self):
        '''Rotate the image's hue by a tiny bit.'''
        new_image = self._original_image.Copy()
        t = time.time()
        new_image.RotateHue((t / 50.) % 1)
        self.bitmap_viewer.set_bitmap(wx.BitmapFromImage(new_image))

        
    def EndModal(self, *args, **kwargs):
        self.timer.Stop()
        wx.Dialog.EndModal(self, *args, **kwargs)

        
        


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This package defines the TreeBrowser class.

See its documentation in the `tree_browser.py` module for more information.
'''

from tree_browser import TreeBrowser

__all__ = ["TreeBrowser"]


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This module defines the SimpackSelectionDialog class.

See its documentation for more info.
'''

import os
import glob
import pkgutil

import wx

from simpack_name_cmp import simpack_name_cmp


class SimpackSelectionDialog(wx.SingleChoiceDialog):
    '''Dialog for selecting a simpack when creating a new gui project.'''
    
    def __init__(self, parent):
        self.make_simpack_list()
        wx.SingleChoiceDialog.__init__(
            self,
            parent,
            'Choose a simulation package for your new simulation:',
            'Choose simulation package',
            self.list_of_simpacks,
            wx.CHOICEDLG_STYLE
        )
        
    def make_simpack_list(self):
        '''Make a list of available simpacks.'''
        import garlicsim_lib.simpacks as simpacks
        self.list_of_simpacks = [
            module_name for (importer, module_name, is_package)
            in pkgutil.iter_modules(simpacks.__path__)
        ]
        self.list_of_simpacks.sort(cmp=simpack_name_cmp)
        

    def get_simpack_selection(self):
        '''Import the selected simpack and return it.'''
        string = self.GetStringSelection()
        result = __import__('garlicsim_lib.simpacks.' + string,
                            fromlist=[''])
        return result






# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the PlaybackControls class.

See its documentation for more info.
'''

from playback_controls import PlaybackControls

__all__ = ["PlaybackControls"]


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the HelpMenu class.

See its documentation for more info.
'''

import wx

from garlicsim_wx.general_misc.cute_menu import CuteMenu

import garlicsim_wx

from online_resources_menu import OnlineResourcesMenu


class HelpMenu(CuteMenu):
    '''Menu for getting help on GarlicSim.'''
    def __init__(self, frame):
        super(HelpMenu, self).__init__()
        self.frame = frame
        self._build()
    
    def _build(self):
        
        frame = self.frame                
        
        
        self.garlicsim_help_button = self.Append(
            wx.ID_HELP_CONTENTS,
            'GarlicSim &Help...\tF1',
            ' Display the help documents for GarlicSim'
        )
        self.garlicsim_help_button.Enable(False)
        
        
        self.welcome_screen_button = self.Append(
            -1,
            '&Welcome screen...',
            ' Show the welcome screen'
        )
        self.welcome_screen_button.Enable(False)
        
                
        self.garlicsim_book_button = self.Append(
            -1,
            'Read the &book, "Introduction to GarlicSim"...',
            ' Open the GarlicSim book, a PDF document'
        )
        self.garlicsim_book_button.Enable(False)
        
        
        self.AppendSeparator()
                
 
        self.online_resources_menu = OnlineResourcesMenu(frame)
        self.online_resources_menu_button = self.AppendMenu(
            -1,
            '&Online resources',
            self.online_resources_menu,
            ' Use resources that require an internet connection'
        )       
        
        
        self.AppendSeparator()
        
                
        self.about_button = self.Append(
            wx.ID_ABOUT,
            '&About GarlicSim...',
            ' Tell me a little bit about the GarlicSim software'
        )
        frame.Bind(
            wx.EVT_MENU,
            lambda event: \
                garlicsim_wx.widgets.misc.AboutDialog(frame).ShowModal(),
            self.about_button
        )
 


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the WorkspaceWidget class.

See its documentation for more info.
'''

import garlicsim_wx
from garlicsim_wx.general_misc.third_party import aui
from garlicsim.general_misc.third_party import abc
from garlicsim.general_misc import string_tools

class WorkspaceWidget(object):
    '''
    Abstract base class for workspace widgets.
    
    A workspace widget is a widget displayed on the Frame of `garlicsim_wx`, and
    is connected to a specific gui project.
    '''
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, frame):
        
        self.frame = frame
        assert isinstance(self.frame, garlicsim_wx.Frame)
        
        self.gui_project = frame.gui_project
        assert isinstance(self.gui_project, garlicsim_wx.GuiProject)
        
        self.aui_manager = frame.aui_manager
        assert isinstance(self.aui_manager, aui.AuiManager)
        
        # I put these asserts mainly for better source assistance in Wing.
        # They may be removed.        

    @classmethod
    def get_uppercase_name(cls):
        '''Get the name of the widget's class in uppercase. Used for title.'''
        return string_tools.camelcase_to_spacecase(cls.__name__).upper()
    
    def get_aui_pane_info(self):
        '''Get the AuiPaneInfo of this widget in the aui manager.'''
        return self.aui_manager.GetPane(self)
        
    


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This is a bootstrap package for garlicsim_wx. See module bootstrap here.
'''

from . import bootstrap


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the StateReprViewer class.

See its documentation for more info.
'''

import wx
from garlicsim_wx.widgets import WorkspaceWidget
import garlicsim.general_misc.dict_tools as dict_tools
from garlicsim_wx.general_misc.flag_raiser import FlagRaiser

__all__ = ["StateReprViewer"]

class StateReprViewer(wx.Panel, WorkspaceWidget):
    '''Widget for showing the repr of the active state.'''
    def __init__(self, frame):
        wx.Panel.__init__(self, frame, size=(300, 300))
        WorkspaceWidget.__init__(self, frame)

        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        
        self.Bind(wx.EVT_PAINT, self.on_paint)        
        
        self.text_ctrl = wx.TextCtrl(
            self,
            style=wx.TE_MULTILINE | wx.NO_BORDER
        )
        
        font_size = 12 if wx.Platform == '__WXMAC__' else 9
        
        font = wx.Font(font_size, wx.DEFAULT, wx.NORMAL, wx.BOLD, False,
                       u'Courier New')
        self.text_ctrl.SetFont(font)
        
        self.sizer_v = wx.BoxSizer(wx.VERTICAL)
        self.sizer_h = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer_v.Add(self.sizer_h, 1, wx.EXPAND)
        self.sizer_h.Add(self.text_ctrl, 1, wx.EXPAND)
        
        self.SetSizer(self.sizer_v)
        self.sizer_v.Layout()
        
        self.state = None
        
        self.needs_recalculation_flag = True
        
        self.needs_update_emitter = \
            self.gui_project.emitter_system.make_emitter(
                inputs=(
                    self.gui_project.active_node_changed_or_modified_emitter,
                    # todo: put the active_state_changed whatever here
                    ),
                outputs=(
                    FlagRaiser(self, 'needs_recalculation_flag',
                               function=self._recalculate, delay=0.03),
                    ),
                name='state_repr_viewer_needs_recalculation',
            )
    

    def _recalculate(self):
        '''Recalculate the widget.'''
        if self.needs_recalculation_flag:
            if self.gui_project:
                active_state = self.gui_project.get_active_state()        
                if active_state:
                    if active_state is not self.state:
                        self.state = active_state
                        state_repr = dict_tools.fancy_string(vars(active_state))
                        self.text_ctrl.SetValue(state_repr)
            self.needs_recalculation_flag = False
        
    def on_paint(self, event):
        '''EVT_PAINT handler.'''
        event.Skip()
        # Notice that we are not checking the `needs_recalculation_flag` here.
        # The FlagRaiser's 30ms delay is small enough, and we don't need to have
        # very fast response time in the state repr viewer, so we can afford to
        # wait another 30ms before an update.
        
         
        
    



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''This package defines wxPython widgets.'''

from workspace_widget import WorkspaceWidget
import misc
import general_misc
import workspace_widgets


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''Images package.'''

import wx
import pkg_resources

from garlicsim.general_misc import caching

from . import images as __images_package
images_package = __images_package.__name__


N_FRAMES = 87
N_BLURRED_GEAR_FRAMES = 6

@caching.cache
def get_image_raw(i):
    '''Get image number `i` of the gear, when 0 <= i <= (N_FRAMES - 1).'''
    
    assert (0 <= i <= N_FRAMES - 1) and isinstance(i, int)
    
    file_name = 'gear%04.d.png' % i
    stream = pkg_resources.resource_stream(images_package, file_name)
    bitmap = wx.BitmapFromImage(
        wx.ImageFromStream(
            stream,
            wx.BITMAP_TYPE_ANY
        )
    )
    
    return bitmap
            

@caching.cache
def get_blur_image_raw(i):
    '''
    Get image number `i` of the blurred gear.

    The higher the `i`, the blurrier is the image.
    
    0 <= i <= (N_BLURRED_GEAR_FRAMES - 1)
    '''
    
    assert isinstance(i, int) and 0 <= i <= (N_BLURRED_GEAR_FRAMES - 1)
    
    file_name = 'blurred_gear_%d.png' % i
    stream = pkg_resources.resource_stream(images_package, file_name)
    bitmap = wx.BitmapFromImage(
        wx.ImageFromStream(
            stream,
            wx.BITMAP_TYPE_ANY
        )
    )

    return bitmap


@caching.cache
def get_blurred_gear_image(i, j):
    '''
    Get image number `i` of the gear, with motion blur number `j`.
    
    0 <= i <= (N_FRAMES - 1)
    0 <= j <= (N_BLURRED_GEAR_FRAMES - 1)
    '''
    image = get_image_raw(i)
    blur = get_blur_image_raw(j)
    bitmap = wx.EmptyBitmap(*get_image_size())
    dc = wx.MemoryDC(bitmap)
    dc.DrawBitmap(image, 0, 0)
    dc.DrawBitmap(blur, 0, 0, True)
    dc.Destroy()
    bitmap.image_raw = image
    bitmap.blur_raw = blur
    return bitmap
    

def get_blurred_gear_image_by_ratio(i, r):
    '''Get the image of the blurred gear by specifying a ratio from 0 to 1.'''
    assert 0 <= r <= 1
    return get_blurred_gear_image(
        i,
        int(round(r * (N_BLURRED_GEAR_FRAMES - 1)))
    )


@caching.cache
def get_image_size():
    '''Get the size of these images here.'''
    return get_image_raw(0).GetSize()
    



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This module defines the SeekBar class.

See its documentation for more info.
'''

import wx
import math

from garlicsim_wx.general_misc.get_lines import get_lines
from garlicsim_wx.general_misc import emitters
from garlicsim_wx.general_misc import wx_tools
from garlicsim_wx.general_misc.flag_raiser import FlagRaiser

from garlicsim_wx.widgets import WorkspaceWidget
import garlicsim
import garlicsim_wx


__all__ = ["SeekBar"]


class SeekBar(wx.Panel, WorkspaceWidget):
    '''
    Seek-bar widget, allowing navigation and visualization of the active path.
    
    The seek-bar is attached to a path. It shows what time period the path
    spans. It shows which node is currently active. It allows to move to any
    other node on the path just by clicking.
    '''
    #todo: show little breaks on the bar where there's a block start/end
    def __init__(self, frame):
        
        wx.Panel.__init__(self, frame, size=(100, 100), style=wx.SUNKEN_BORDER)
        WorkspaceWidget.__init__(self, frame)
        
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        
        self.Bind(wx.EVT_PAINT, self.on_paint)
        self.Bind(wx.EVT_SIZE, self.on_size)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.on_mouse_event)

        self.zoom = 1.
        self.start = 0.

        self.screenify = lambda x: (x-self.start)*self.zoom
        '''Translate from time point to on-screen coordinate.'''
        
        self.unscreenify = lambda x: (x/self.zoom)+self.start
        '''Translate from on-screen coordinate to time point.'''
        
        self.menu = garlicsim_wx.general_misc.cute_menu.CuteMenu.add_menus(
            [garlicsim_wx.misc.menu_bar.node_menu.NodeMenu(self.frame),
             garlicsim_wx.misc.menu_bar.block_menu.BlockMenu(self.frame)]
        )

        self.was_playing_before_mouse_click = None
        self.was_playing_before_mouse_click_but_then_paused_and_mouse_left = None
        self.active_triangle_width = 13 # Must be odd number

        self.view_changed_flag = False
        self.active_node_changed_or_modified_flag = False
        self.path_contents_changed_flag = False
        
        self.view_change_emitter = emitters.Emitter()
        self.gui_project.active_node_changed_or_modified_emitter.add_output(
            FlagRaiser(self, 'active_node_changed_or_modified_flag')
        )
        self.gui_project.path_contents_changed_emitter.add_output(
            FlagRaiser(self, 'path_contents_changed_flag')
        )


    def on_paint(self, event):
        '''Refresh the seek-bar.'''
        event.Skip()
        

        self.view_changed_flag = False
        self.active_node_changed_or_modified_flag = False
        self.path_contents_changed_flag = False
        # todo: now we just lower these flags retardedly, in future there will
        # be __recalculate
        
        occupied_region = wx.Region()

        if (self.gui_project is None) or (self.gui_project.path is None):
            return
        
        (w, h) = self.GetSize()
        start = self.start
        end = self.start + (w / self.zoom)
        dc = wx.BufferedPaintDC(self)

        dc.SetBackground(wx_tools.get_background_brush())
        dc.Clear()
        
        #dc.DrawRectangle(3,3,50,90)
        if self.gui_project.path is not None: # Draw rect for renedered segment
            seg = self.gui_project.path.get_existing_time_segment(start, end)
            if seg is not None:
                dc.SetPen(wx.Pen('#000000'))
                dc.SetBrush(wx.Brush('#FFFFB8'))
                sseg = [self.screenify(thing) for thing in seg]
                dc.DrawRectangle(sseg[0], 0, sseg[1]-sseg[0], h-4)
                occupied_region = wx.Region(sseg[0] + 1, 1,
                                            sseg[1] - sseg[0] - 2,
                                            h - 6)

        active = self.gui_project.active_node
        if active is not None:
            active_start = active.state.clock
            try:
                after_active = self.gui_project.path.next_node(active)
                active_end = after_active.state.clock
            except garlicsim.data_structures.path.PathOutOfRangeError:
                after_active = None
                active_end = active_start
            active_inside = False
            screen_active_start = start
            screen_active_end = end

            if start <= active_start <= end:
                active_inside = True
                screen_active_start = self.screenify(active_start)

            if start <= active_end <= end:
                active_inside = True
                screen_active_end = self.screenify(active_end)


            dc.SetBrush(wx.Brush('#FF9933'))
            dc.SetPen(wx.Pen('#000000', 1, wx.TRANSPARENT))
            if active_inside is True:
                dc.DrawRectangle(math.floor(screen_active_start),
                                 1,
                                 math.ceil(screen_active_end-screen_active_start),
                                 h-6)
                triangle_half_width = math.ceil(self.active_triangle_width / 2.0)
                dc.SetClippingRegionAsRegion(occupied_region)
                dc.DrawPolygon(
                        ((screen_active_start - triangle_half_width, h - 5),
                        (screen_active_start + triangle_half_width, h - 5),
                        (screen_active_start, h - 5 - triangle_half_width))
                    )
                dc.DestroyClippingRegion()




        # Draw ruler
        min = 15
        temp = math.ceil(math.log10(min / self.zoom))
        bigliners = get_lines(start, end, temp+1)
        presmallliners = get_lines(start, end, temp)
        smallliners = []
        for thing in presmallliners:
            if bigliners.count(thing) == 0:
                smallliners.append(thing)

        self.draw_small_numbers(dc, smallliners)
        self.draw_big_numbers(dc, bigliners)
        
        dc.Destroy()


    def draw_small_numbers(self, dc, numbers):
        dc.SetPen(wx.Pen('#000000'))
        dc.SetFont(
            wx.Font(9,
                    wx.FONTFAMILY_DEFAULT,
                    wx.FONTSTYLE_NORMAL,
                    wx.FONTWEIGHT_NORMAL,
                    False,
                    'Courier 10 Pitch')
        )
        
        for number in numbers:
            dc.DrawLine(number, 0, number, 6)
            width, height = dc.GetTextExtent(str(number))
            dc.DrawText(str(number), (number - width / 2), 8)

    def draw_big_numbers(self, dc, numbers):
        dc.SetPen(wx.Pen('#000000'))
        dc.SetFont(
            wx.Font(9,
                    wx.FONTFAMILY_DEFAULT,
                    wx.FONTSTYLE_NORMAL,
                    wx.FONTWEIGHT_BOLD,
                    False,
                    'Courier 10 Pitch')
        )
        
        for number in numbers:
            dc.DrawLine(number, 0, number, 9)
            width, height = dc.GetTextExtent(str(number))
            dc.DrawText(str(number), (number - width / 2), 12)


    def on_mouse_event(self, e):
        #todo: should catch drag to outside of the window        
        # todo: use EVT_CONTEXT_MENU, in tree browser and others too
        if e.RightDown():
            self.gui_project.stop_playing()

            reselect_node = False
            new_thing = e.GetPositionTuple()[0]
            if self.gui_project.active_node is None:
                reselect_node=True
            else:
                thing = self.screenify(self.gui_project.active_node.state.clock)
                if abs(thing - new_thing) >= 8:
                    reselect_node = True

            if reselect_node is True:
                
                new_node = self.gui_project.path.get_node_occupying_timepoint \
                         (self.unscreenify(new_thing))
                
                if new_node is not None:
                    self.gui_project.set_active_node(new_node, modify_path=False)

            if self.gui_project.active_node is not None:
                self.gui_project.frame.Refresh()
                self.PopupMenu(self.menu, e.GetPosition())



        if e.LeftDClick():
            self.gui_project.toggle_playing()
            
        if e.LeftDown():# or e.RightDown():
            thing = e.GetPositionTuple()[0]
            node = self.gui_project.path.get_node_occupying_timepoint \
                 (self.unscreenify(thing))

            self.was_playing_before_mouse_click = self.gui_project.is_playing
            if self.was_playing_before_mouse_click:
                self.gui_project.stop_playing()

            if node is not None:
                self.gui_project.set_active_node(node, modify_path=False)


        if e.LeftIsDown():
            thing = e.GetPositionTuple()[0]
            node = self.gui_project.path.get_node_occupying_timepoint \
                 (self.unscreenify(thing))
            if node is not None:
                self.gui_project.set_active_node(node, modify_path=False)
                
        if e.LeftUp():
            if self.was_playing_before_mouse_click:
                self.gui_project.start_playing()
                self.was_playing_before_mouse_click = False
                
        if e.Leaving():
            if self.was_playing_before_mouse_click:
                self.gui_project.start_playing()
                self.was_playing_before_mouse_click = False
                self.was_playing_before_mouse_click_but_then_paused_and_mouse_left = True
                
        if e.Entering():
            if self.was_playing_before_mouse_click_but_then_paused_and_mouse_left:
                self.gui_project.stop_playing()
                self.was_playing_before_mouse_click = True
                self.was_playing_before_mouse_click_but_then_paused_and_mouse_left = False


    def on_size(self, e):
        self.Refresh()
        if e is not None:
            e.Skip()



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Widgets used in the workspace.
'''

from state_repr_viewer import StateReprViewer
from shell import Shell
from seek_bar import SeekBar
from tree_browser import TreeBrowser
from playback_controls import PlaybackControls



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
A wxPython-based GUI for garlicsim.

The final goal of this project is to become a fully-fledged application for
working with simulations, friendly enough that it may be used by
non-programmers.

This program is intended for Python versions 2.5 and 2.6.
'''

import bootstrap

import sys
import os.path

import wx

import misc
from frame import Frame
from gui_project import GuiProject
from app import App


__all__ = ['Frame', 'GuiProject', 'start']


__version__ = '0.5'


def start():
    '''Start the GUI.'''
    
    # The first argument is the path of the script (or the executable if we're
    # frozen), and that should be cut off:    
    args = sys.argv[1:]
    
    # todo: Consider removing the args we can understand from sys.argv, so
    # program inside will not be confused by them.
        
    new_gui_project_simpack_name = None
    load_gui_project_file_path = None
    
    if args:
        arg = args[0]        
        if arg.startswith('__garlicsim_wx_new='):
            new_gui_project_simpack_name = arg[19:]
        elif os.path.isfile(arg):
            load_gui_project_file_path = arg
    
    app = App(new_gui_project_simpack_name=new_gui_project_simpack_name,
              load_gui_project_file_path=load_gui_project_file_path)
    
    app.MainLoop()
    
    
if __name__ == "__main__":
    
    start()


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the simpack_name_cmp function.

See its documentation for more info.
'''



__all__ = ['simpack_name_cmp']

def _leading_underscore_count(string):
    '''Count the number of leading underscores that a string has.'''
    result = 0
    for char in string:
        if char == '_':
            result += 1
        else:
            break
    return result

def _leading_underscore_cmp(a, b):
    '''Compare which string has more leading underscores.'''
    return cmp(
        _leading_underscore_count(a),
        _leading_underscore_count(b)
    )

def simpack_name_cmp(a, b):
    '''Compare which simpack has a "higher" name.'''
    assert isinstance(a, basestring) and isinstance(b, basestring)
    first_result = _leading_underscore_cmp(a, b)
    if first_result != 0:
        return first_result
    return cmp(a, b)


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This module defines the FlagRaiser class.

See its documentation for more info.
'''

import wx

from garlicsim_wx.general_misc import cute_timer


class FlagRaiser(object): # todo: rename?
    '''When called, raises a flag of a window and then calls some function.'''
    def __init__(self, window, attribute_name=None, function=None, delay=None):
        '''
        Construct the flag raiser.
        
        `window` is the window we're acting on. `attribute_name` is the name of
        the flag that we set to True. `function` is the function we call after
        we set the flag. Default for `function` is `window.Refresh`.
        
        If we get a `delay` argument, then we don't call the function
        immediately, but wait for `delay` time, specified as seconds, then call
        it. If this flag raiser will be called again while the timer's on, it
        will not cause another function calling.
        '''
        assert isinstance(window, wx.Window)

        self.window = window
        '''The window that the flag raiser is acting on.'''
        
        self.attribute_name = attribute_name
        '''The name of the flag that this flag raiser raises.'''
        
        self.function = function or window.Refresh
        '''The function that this flag raiser calls after raising the flag.'''
        
        self.delay = delay
        '''The delay, in seconds, that we wait before calling the function.'''
        
        if delay is not None:
            
            self._delay_in_ms = delay * 1000
            '''The delay in milliseconds.'''
            
            self.timer = cute_timer.CuteTimer(self.window)
            '''The timer we use to call the function.'''
            
            self.window.Bind(wx.EVT_TIMER, self.on_timer, self.timer)

            
    def __call__(self):
        '''Raise the flag and call the function. (With delay if we set one.)'''
        if self.attribute_name:
            setattr(self.window, self.attribute_name, True)
        if self.delay is None:
            self.function()
        else: # self.delay is a positive number
            if not self.timer.IsRunning():
                self.timer.Start(self._delay_in_ms, oneShot=True)
                
    def on_timer(self, event):
        '''EVT_TIMER event handler.'''
        if getattr(self.window, self.attribute_name) is True:
            self.function()


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the Knob class.

See its documentation for more info.
'''

from __future__ import division

import math

import wx
import pkg_resources

from garlicsim.general_misc import math_tools
from garlicsim_wx.general_misc import wx_tools
from garlicsim_wx.general_misc import cursor_collection
from garlicsim.general_misc import binary_search
from garlicsim.general_misc import cute_iter_tools

from snap_map import SnapMap

from . import images as __images_package
images_package = __images_package.__name__


class Knob(wx.Panel):
    '''
    A knob that sets a real value between -Infinity and Infinity.
    
    (Not really touching infinity.)
    
    By turning the knob with the mouse, the user changes a floating point
    variable.


    There are three "scales" that one should keep in mind when working with Knob:
    
    1. The "value" scale, which is the value that the actual final variable
       gets. It spans from -Infinity to Infinity.
    
    2. The "angle" scale, which is the angle in which the knob appears on
        the screen. It span from (-(5/6) * pi) to ((5/6) * pi).
    
    3. As a more convenient mediator between them there's the "ratio" scale,
       which spans from -1 to 1, and is mapped linearly to "angle".
    
    
    The knob has snap points that can be modified with `.set_snap_point` and
    `.remove_snap_point`. These are specified by value.
    '''
    # todo future: make key that disables snapping while dragging
    # todo: consider letting the knob turn just a bit slower near the edges.
    # todo: currently forcing size to be constant, in future allow changing
    def __init__(self, parent, getter, setter, *args, **kwargs):
        '''
        Construct the knob.
        
        `getter` is the getter function used to get the value of the variable.
        `setter` is the setter function used to set the value of the variable.
        
        Note that you can't give a size argument to knob, it is always created
        with a size of (29, 29).
        '''
        
        assert 'size' not in kwargs
        kwargs['size'] = (29, 29)
        
        assert callable(setter) and callable(getter)
        self.value_getter, self.value_setter = getter, setter
        
        wx.Panel.__init__(self, parent, *args, **kwargs)
        
        self.original_bitmap = wx.BitmapFromImage(
            wx.ImageFromStream(
                pkg_resources.resource_stream(images_package, 'knob.png'),
                wx.BITMAP_TYPE_ANY
            )
        )
        
        self.Bind(wx.EVT_PAINT, self.on_paint)
        self.Bind(wx.EVT_SIZE, self.on_size)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.on_mouse)
        # self.Bind(wx.EVT_ERASE_BACKGROUND, self.on_erase)
        
        self.SetCursor(cursor_collection.get_open_grab())
        
        
        self._knob_house_brush = wx.Brush(wx.Color(0, 0, 0))
        '''Brush used to paint the circle around the knob.'''
        
        self.current_angle = 0
        '''The current angle of the knob.'''
        
        self.current_ratio = 0
        '''The current ratio of the knob.'''
        
        self.sensitivity = 25
        '''
        The knob's sensitivity.
        
        Higher values will cause faster changes in value when turning the knob.
        '''
        
        self.angle_resolution = math.pi / 180
        '''The minimal change in angle that will warrant a repaint.'''
        
        self.snap_points = []
        '''An ordered list of snap points, specified by value.'''
        
        self.base_drag_radius = 50
        '''
        The base drag radius, in pixels.
        
        This number is the basis for calculating the height of the area in which
        the user can play with the mouse to turn the knob. Beyond that area the
        knob will be turned all the way to one side, and any movement farther
        will have no effect.
        
        If there are no snap points, the total height of that area will be `2 *
        self.base_drag_radius`.
        '''
        
        self.snap_point_drag_well = 20
        '''
        The height of a snap point's drag well, in pixels.
        
        This is the height of the area on the screen in which, when the user
        drags to it, the knob will have the value of the snap point.
        
        The bigger this is, the harder the snap point "traps" the mouse.
        '''
            
        self.being_dragged = False
        '''Flag saying whether the knob is currently being dragged.'''
        
        self.snap_map = None
        '''
        The current snap map used by the knob.
        
        See documentation of SnapMap for more info.
        '''
        
        self.needs_recalculation_flag = True
        '''Flag saying whether the knob needs to be recalculated.'''
        
        self._recalculate()

    
    def _angle_to_ratio(self, angle):
        '''Convert from angle to ratio.'''
        return angle / (math.pi * 5 / 6)

    def _ratio_to_value(self, ratio):
        '''Convert from ratio to value.'''
        return self.sensitivity * \
               math_tools.sign(ratio) * \
               (4 / math.pi**2) * \
               math.log(math.cos(ratio * math.pi / 2))**2
        
    def _value_to_ratio(self, value):
        '''Convert from value to ratio.'''
        return math_tools.sign(value) * \
               (2 / math.pi) * \
               math.acos(
                   math.exp(
                       - (math.pi * math.sqrt(abs(value))) / \
                       (2 * math.sqrt(self.sensitivity))
                   )
               )

    def _ratio_to_angle(self, ratio):
        '''Convert from ratio to angle.'''
        return ratio * (math.pi * 5 / 6)
    
    def _get_snap_points_as_ratios(self):
        '''Get the list of snap points, but as ratios instead of as values.'''
        return [self._value_to_ratio(value) for value in self.snap_points]
    
    def set_snap_point(self, value):
        '''Set a snap point. Specified as value.'''
        # Not optimizing with the sorting for now
        self.snap_points.append(value)
        self.snap_points.sort()
    
    def remove_snap_point(self, value):
        '''Remove a snap point. Specified as value.'''
        self.snap_points.remove(value)
        
    def _recalculate(self):
        '''
        Recalculate the knob, changing its angle and refreshing if necessary.
        '''
        value = self.value_getter()
        self.current_ratio = self._value_to_ratio(value)
        angle = self._ratio_to_angle(self.current_ratio)
        d_angle = angle - self.current_angle
        if abs(d_angle) > self.angle_resolution:
            self.current_angle = angle
            self.Refresh()
        self.needs_recalculation_flag = False
    
    def on_paint(self, event):
        '''EVT_PAINT handler.'''
        
        # Not checking for recalculation flag, this widget is not real-time
        # enough to care about the delay.
        
        dc = wx.BufferedPaintDC(self)
        
        dc.SetBackground(wx_tools.get_background_brush())
        dc.Clear()
        
        w, h = self.GetClientSize()
        
        gc = wx.GraphicsContext.Create(dc)

        gc.SetPen(wx.TRANSPARENT_PEN)
        gc.SetBrush(self._knob_house_brush)
        
        assert isinstance(gc, wx.GraphicsContext)
        gc.Translate(w/2, h/2)
        gc.Rotate(self.current_angle)
        gc.DrawEllipse(-13.5, -13.5, 27, 27)
        gc.DrawBitmap(self.original_bitmap, -13, -13, 26, 26)
        
        #gc.DrawEllipse(5,5,2,2)
        #gc.DrawEllipse(100,200,500,500)
        
    def on_size(self, event):
        '''EVT_SIZE handler.'''
        event.Skip()
        self.Refresh()
      
    def on_mouse(self, event):
        '''EVT_MOUSE_EVENTS handler.'''
        # todo: maybe right click should give context menu with 'Sensitivity...'
        # todo: make check: if left up and has capture, release capture

        self.Refresh()
        
        (w, h) = self.GetClientSize()
        (x, y) = event.GetPositionTuple()
        
        
        if event.LeftDown():
            self.being_dragged = True
            self.snap_map = SnapMap(
                snap_point_ratios=self._get_snap_points_as_ratios(),
                base_drag_radius=self.base_drag_radius,
                snap_point_drag_well=self.snap_point_drag_well,
                initial_y=y,
                initial_ratio=self.current_ratio
            )
            
            self.SetCursor(cursor_collection.get_closed_grab())
            # SetCursor must be before CaptureMouse because of wxPython/GTK
            # weirdness
            self.CaptureMouse()
            
            return
        
        if event.LeftIsDown() and self.HasCapture():
            ratio = self.snap_map.y_to_ratio(y)
            value = self._ratio_to_value(ratio)
            self.value_setter(value)
            
                
        if event.LeftUp():
            # todo: make sure that when leaving
            # entire app, things don't get fucked
            if self.HasCapture():
                self.ReleaseMouse()
            # SetCursor must be after ReleaseMouse because of wxPython/GTK
            # weirdness
            self.SetCursor(cursor_collection.get_open_grab())
            self.being_dragged = False
            self.snap_map = None
            
            
        return
    
    """
    def on_erase(self, event):
        pass
    """
        

        
        
        
        
        
        


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the GuiProject class.

See its documentation for more info.
'''

from __future__ import with_statement

import warnings
import copy
import functools
import Queue
import time

import wx
import wx.lib.scrolledpanel
import wx.py.shell

from garlicsim.general_misc import queue_tools
from garlicsim.general_misc import dict_tools
from general_misc.stringsaver import s2i,i2s
from garlicsim.general_misc.infinity import Infinity
from garlicsim.general_misc import binary_search
from garlicsim_wx.general_misc import thread_timer
from garlicsim_wx.general_misc import wx_tools

import garlicsim
from garlicsim.asynchronous_crunching import crunchers
import garlicsim_wx
from garlicsim_wx.general_misc import emitters
        

class GuiProject(object):
    '''Encapsulates a project for use with a wxPython interface.'''
    
    def __init__(self, simpack, frame, project=None, virgin=True):
        '''
        Construct the gui project.
        
        `simpack` is the simpack (or grokker) to use. `frame` is the frame in
        which this gui project will live. `virgin` should remain True in normal
        usage, it's set to False only when loading a GuiProject from file.
        '''
        # This is broken down into a few parts.
        self.__init_general(simpack, frame, project)
        self.__init_gui()
        if virgin:
            self.__init_virgin()

    
    @staticmethod
    def load_from_vars(frame, picklable_vars):
        '''Take vars that were just unpickled and build a GuiProject from them.'''

        simpack, project = (
            picklable_vars['simpack'],
            picklable_vars['project']
        )
        
        gui_project = GuiProject(simpack, frame, project, virgin=False)
        
        for (key, value) in picklable_vars.iteritems():
            setattr(gui_project, key, value)
        
        return gui_project
    
        
    def __init_general(self, simpack, frame, project=None):
        '''General initialization.'''
        
        self.frame = frame
        '''The frame that this gui project lives in.'''
        
        if isinstance(simpack, garlicsim.misc.SimpackGrokker):
            simpack_grokker = simpack            
            simpack = simpack_grokker.simpack
        else:
            simpack_grokker = garlicsim.misc.SimpackGrokker(simpack)
            
            
        self.simpack = simpack
        '''The simpack used for this gui project.'''
        
        self.simpack_grokker = simpack_grokker
        '''The simpack grokker used for this gui project.'''
        
        self.simpack_wx_grokker = garlicsim_wx.misc.SimpackWxGrokker(simpack)
        '''The simpack_wx used for this gui project.'''
        
        self.project = project or garlicsim.Project(simpack_grokker)
        '''The project encapsulated in this gui project.'''
        

        ### Choosing a Cruncher class: ########################################
        
        if self.project.simpack_grokker.history_dependent is False and \
           self.project.simpack_grokker.settings.FORCE_CRUNCHER is None and \
           'CruncherProcess' in vars(crunchers):
            
            self.project.crunching_manager.Cruncher = crunchers.CruncherProcess
        
        #######################################################################
            
        self.path = None
        '''The active path.'''

        self.active_node = None
        '''The node that is currently displayed onscreen.'''

        self.is_playing = False
        '''Says whether the simulation is currently playing.'''
        
        self.infinity_job = None
        '''
        The job of the playing leaf, which should be crunched to infinity.
        '''
        
        self.default_buffer = 100 # Should be a mechanism for setting that
        '''The default clock buffer to crunch from an active node.'''

        self.timer_for_playing = thread_timer.ThreadTimer(self.frame)
        '''Contains the wx.Timer object used when playing the simulation.'''
        
        self.frame.Bind(thread_timer.EVT_THREAD_TIMER, self.__play_next,
                        self.timer_for_playing)

        self.defacto_playing_speed = 4
        '''The playing speed that we are actually playing in.'''
        
        self.official_playing_speed = 4
        '''
        The playing speed that we're "officially" playing in.
        
        The defacto playing speed may deviate from this.
        '''
        
        self.standard_playing_speed = 4
        '''The reference playing speed. This speed is considered "normal".'''
        
        self.last_tracked_real_time = None
        '''
        The last tracked time point (real time, not simulation), for playback.
        '''
        
        self.pseudoclock = 0
        '''
        The current pseudoclock.
        
        The pseudoclock is *something like* the clock of the current active
        node. But not exactly. We're letting the pseudoclock slide more smoothly
        from one node to its neighbor, instead of jumping. This is in order to
        make some things smoother in the program. This is also why it's called
        "pseudo".
        '''

        self.__init_emitters()
        self.__init_menu_enablings()
        
        self.emitter_system.top_emitter.emit()
        # Just for good measure, jiggle all the widgets up.
        
        
    def __init_emitters(self):
        '''Create an emitter system and a bunch of emitters.'''
        
        # todo: not clear that `tree_modified_emitter` means that only data
        # changed and not structure.
        
        self.emitter_system = emitters.EmitterSystem()
                
        with self.emitter_system.freeze_cache_rebuilding:
        
            es = self.emitter_system
            
            self.tree_modified_emitter = es.make_emitter(
                name='tree_modified',
            )
            self.tree_modified_on_path_emitter = es.make_emitter(
                outputs=(self.tree_modified_emitter,),
                name='tree_modified_on_path',
            )
    
            self.tree_modified_not_on_path = es.make_emitter(
                outputs=(self.tree_modified_emitter,),
                name='tree_modified_not_on_path',
            )
            
            self.tree_modified_at_unknown_location_emitter = es.make_emitter(
                outputs=(
                    self.tree_modified_on_path_emitter,
                    self.tree_modified_not_on_path,
                ),
                name='tree_modified_at_unknown_location',
            )
            
            self.tree_structure_modified_emitter = es.make_emitter(
                outputs=(self.tree_modified_emitter,),
                name='tree_structure_modified',
            )
            
            self.tree_structure_modified_on_path_emitter = es.make_emitter(
                outputs=(
                    self.tree_modified_on_path_emitter,
                    self.tree_structure_modified_emitter
                ),
                name='tree_structure_modified_on_path',
            )
            self.tree_structure_modified_not_on_path_emitter = es.make_emitter(
                outputs=(
                    self.tree_modified_not_on_path,
                    self.tree_structure_modified_emitter
                ),
                name='tree_structure_modified_not_on_path',
            )
            self.tree_structure_modified_at_unknown_location_emitter = \
                es.make_emitter(
                outputs=(
                    self.tree_structure_modified_on_path_emitter,
                    self.tree_structure_modified_not_on_path_emitter,
                    self.tree_modified_at_unknown_location_emitter
                ),
                name='tree_structure_modified_at_unknown_location',
            )
            
    
            self.pseudoclock_modified_emitter = es.make_emitter(
                name='pseudoclock_modified'
            )
    
            self.active_node_changed_emitter = es.make_emitter(
                name='active_node_changed'
            )
            # todo: should possibly take input from pseudoclock_modified_emitter
            
            self.active_node_modified_emitter = es.make_emitter(
                name='active_node_modified'
            )
            
            self.active_node_changed_or_modified_emitter = es.make_emitter(
                name='active_node_changed_or_modified',
                inputs=(
                    self.active_node_changed_emitter,
                    self.active_node_modified_emitter,
                )
            )
            
            self.path_changed_emitter = es.make_emitter(
                name='path_changed'
            )
            
            self.path_contents_changed_emitter = es.make_emitter(
                inputs=(
                    self.path_changed_emitter,
                    self.tree_modified_on_path_emitter
                ),
                name='path_contents_changed',
            )
            
            self.playing_toggled_emitter = es.make_emitter(
                name='playing_toggled',
            )
            self.playing_started_emitter = es.make_emitter(
                outputs=(self.playing_toggled_emitter,),
                name='playing_started',
            )
            self.playing_stopped_emitter = es.make_emitter(
                outputs=(self.playing_toggled_emitter,),
                name='playing_stopped',
            )
    
            self.official_playing_speed_modified_emitter = es.make_emitter(
                    outputs=(self.update_defacto_playing_speed,),
                    name='official_playing_speed_modified',
                )
            
            self.active_node_finalized_emitter = es.make_emitter(
                inputs=(self.active_node_modified_emitter,),
                outputs=(self.tree_modified_on_path_emitter,), # todo: correct?
                name='active_node_finalized',
            )
            
            #todo: maybe need an emitter for when editing a state?
            
            ###################################################################
                        
            self.all_menus_need_recalculation_emitter = es.make_emitter(
                outputs=(self.frame._recalculate_all_menus,),
                name='all_menus_need_recalculation_emitter'
            )

    
    def __init_menu_enablings(self):
        '''Connect the functions that (en/dis)able menus to the emitters.'''
        for menu in [self.frame.menu_bar.node_menu,
                     self.frame.menu_bar.block_menu]:
            
            self.active_node_changed_or_modified_emitter.add_output(
                menu._recalculate
            )
        
            
    def __init_gui(self):
        '''
        Initialization related to the widgets which make up the gui project.
        '''
        
        self.frame.Bind(wx.EVT_MENU, self.fork_by_editing,
                         id=s2i("Fork by editing"))
        self.frame.Bind(wx.EVT_MENU, self.fork_by_crunching,
                         id=s2i("Fork by crunching"))
        
        
    def __init_virgin(self):
        '''
        Initialization done when gui project is actually created, not loaded.
        '''
        wx.CallAfter(self.make_state_creation_dialog)
        

    def set_path(self, path):
        '''Set the path.'''
        self.path = path
        self.path_changed_emitter.emit()
        
        
    def set_official_playing_speed(self, value):
        '''Set the official playing speed.'''
        self.official_playing_speed = value
        self.official_playing_speed_modified_emitter.emit()

        
    def _set_pseudoclock(self, value):
        '''Set the pseudoclock. Internal use.'''
        if self.pseudoclock != value:
            self.pseudoclock = value
            self.pseudoclock_modified_emitter.emit()

            
    def set_pseudoclock(self, desired_pseudoclock,
                        rounding=binary_search.LOW_OTHERWISE_HIGH):
        '''
        Attempt to set the pseudoclock to a desired value.
        
        If value is outside the range of the current path, you'll get the clock
        of the closest edge node.
        
        The active node will be changed to one which is close to the desired
        pseudoclock. In `rounding` use `binary_search.LOW_OTHERWISE_HIGH` to get
        the node just below, or `binary_search.HIGH_OTHERWISE_LOW` to get the
        node just above.
        
        See documentation for these two options for more details.
        '''
        # todo: check that everything that should use this does use this

        assert rounding in (binary_search.LOW_OTHERWISE_HIGH,
                            binary_search.HIGH_OTHERWISE_LOW)
        
        both_nodes = self.path.get_node_by_clock(desired_pseudoclock,
                                                 rounding=binary_search.BOTH)
        
        binary_search_profile = binary_search.BinarySearchProfile(
            self.path, 
            lambda node: node.state.clock,
            desired_pseudoclock,
            both_nodes
        )
        
        node = binary_search_profile.results[rounding]
        
        if node is None:
            return # todo: Not sure if I should raise something
        
        self._set_active_node(node)
        
        if binary_search_profile.is_surrounded:
            self._set_pseudoclock(desired_pseudoclock)
        else:
            self._set_pseudoclock(node.state.clock)
        
        self.project.ensure_buffer(node, clock_buffer=self.default_buffer)

        
    def round_pseudoclock_to_active_node(self):
        '''Set the value of the pseudoclock to the clock of the active node.'''
        self._set_pseudoclock(self.active_node.state.clock)

        
    def update_defacto_playing_speed(self):
        '''Update the defacto playing speed to the official playing speed.'''
        # In the future this will check if someone's temporarily tweaking the
        # defacto speed, and let that override.
        self.defacto_playing_speed = self.official_playing_speed
        
    
    def make_state_creation_dialog(self):
        '''Create a dialog for creating a root state.'''
        Dialog = self.simpack_wx_grokker.settings.STATE_CREATION_DIALOG
        dialog = Dialog(self.frame)
        state = dialog.start()
        if state:
            root = self.project.root_this_state(state)
            self.tree_structure_modified_not_on_path_emitter.emit()
            self.set_active_node(root)
        

    def get_active_state(self):
        '''Get the active state, i.e. the state of the active node.'''
        return self.active_node.state if self.active_node is not None else None
                

    
    def _set_active_node(self, node):
        '''Set the active node, displaying it onscreen. Internal use.'''
        if self.active_node is not node:
            self.active_node = node        
            self.active_node_changed_emitter.emit()

        
    
    def set_active_node(self, node, modify_path=True):
        '''
        Set the active node, displaying it onscreen.
        
        This will change the pseudoclock to the clock of the node.
        
        if `modify_path` is True, the method will modify the path to go through
        the node, if it doesn't already.        
        '''
        self.project.ensure_buffer(node, clock_buffer=self.default_buffer)
        
        if self.active_node is node:
            return
        
        was_playing = self.is_playing # todo: consider cancelling this
        if self.is_playing: self.stop_playing()
        
        self._set_active_node(node)

        self._set_pseudoclock(node.state.clock)
        
        if was_playing:
            self.start_playing()
        if modify_path:
            self.__modify_path_to_include_active_node()
            
        if modify_path and was_playing:
            if self.infinity_job:
                self.infinity_job.crunching_profile.clock_target = \
                    self.infinity_job.node.state.clock + self.default_buffer
                self.infinity_job = self.project.ensure_buffer_on_path(node,
                                                                       self.path,
                                                                       Infinity)   
        

        
    def __modify_path_to_include_active_node(self):
        '''Ensure that self.path includes the active node.'''
        if self.path is None:
            self.set_path(self.active_node.make_containing_path())
        else:
            self.path.modify_to_include_node(self.active_node)
            
        self.path_changed_emitter.emit()


    def start_playing(self):
        '''Start playback of the simulation.'''
        if self.is_playing:
            return
        if self.active_node is None:
            return

        self.is_playing = True
        
        
        self.infinity_job = \
            self.project.ensure_buffer_on_path(self.active_node, self.path,
                                                 Infinity)
        
        self.timer_for_playing.Start(1000//25)
        
        assert self.last_tracked_real_time == None
        self.round_pseudoclock_to_active_node()
        self.last_tracked_real_time = time.time()        
        self.playing_started_emitter.emit()
        
        # todo: maybe should start a call of __play_next right here, to save
        # 25ms delay on the first frame?
        


    def stop_playing(self):
        '''Stop playback of the simulation.'''
        
        if self.is_playing is False:
            return

        try:
            self.timer_for_playing.Stop()
        except Exception: # todo: Find out the type
            pass
        
        self.is_playing = False
        
        if self.infinity_job:
            self.infinity_job.crunching_profile.clock_target = \
                self.infinity_job.node.state.clock + self.default_buffer
        
        self.last_tracked_real_time = None
        self.round_pseudoclock_to_active_node()
        self.project.ensure_buffer(self.active_node, self.default_buffer)
        
        self.playing_stopped_emitter.emit()


    def editing_state(self):
        '''
        Get a state suitable for editing.
        
        If the current active node is "still in editing", returns its state. If
        not, forks the tree with the active node as a template and returns the
        newly created state.
        '''
        node = self.active_node
        state = node.state
        if (node.touched is False) or (node.still_in_editing is False):
            new_node = self.fork_by_editing()
            return new_node.state
        else:
            return state

        
    def toggle_playing(self):
        '''Toggle the onscreen playback of the simulation.'''
        return self.stop_playing() if self.is_playing else self.start_playing()
    
        
    def __play_next(self, event=None):
        '''
        Show the next node onscreen.
        
        This method is called repeatedly when playing the simulation.
        '''
        if self.is_playing is False: return

        current_real_time = time.time()
        real_time_elapsed = (current_real_time - self.last_tracked_real_time)
        desired_pseudoclock = \
            self.pseudoclock + \
            (real_time_elapsed * self.defacto_playing_speed)
        

        rounding = binary_search.LOW_OTHERWISE_HIGH \
                 if self.defacto_playing_speed > 0 \
                 else binary_search.HIGH_OTHERWISE_LOW
        
        self.set_pseudoclock(desired_pseudoclock, rounding)

        self.last_tracked_real_time = current_real_time
        

    def fork_by_crunching(self, e=None):
        '''
        Fork the simulation from the active node.
        
        Used for forking the simulation without modifying any states. Creates
        a new node from the active node via natural simulation.
        '''
        #todo: maybe not let to do it from unfinalized touched node?
        
        node = self.active_node
        self.project.begin_crunching(self.active_node, self.default_buffer)


    def fork_by_editing(self, e=None):
        '''
        Fork the simulation from the active node by editing.
        
        Returns the new node.
        '''
        # todo: event argument is bad, in other places too
        # todo: maybe not restrict it to "from_active_node"?
        new_node = \
            self.project.tree.fork_to_edit(template_node=self.active_node)
        self.tree_structure_modified_on_path_emitter.emit()
        self.set_active_node(new_node)
        return new_node


    def sync_crunchers(self):
        '''
        Take work from the crunchers, and give them new instructions if needed.
        
        (This is a wrapper for Project.sync_crunchers() with some gui-related
        additions.)
        
        Talks with all the crunchers, takes work from them for implementing
        into the tree, retiring crunchers or recruiting new crunchers as
        necessary.

        Returns the total amount of nodes that were added to the tree in the
        process.
        '''
        
        # This method basically has two tasks. The first one is to take work
        # from the crunchers and retire/recruit/redirect them as necessary. This
        # is done by `Project.sync_crunchers`, which we call here, so that's not
        # the tricky part here.
        #
        # The second task is the tricky part. We want to know just how much the
        # tree was modified during this action. And the tricky thing is that
        # `Project.sync_crunchers` won't do that for us, so we're going to have
        # to try to deduce how much the tree modified ourselves, just by looking
        # at the `jobs` list before and after calling `Project.sync_crunchers`.
        #
        # And when I say "to know how much the tree modified", I mean mainly to
        # know if the modification is a structural modification, or just some
        # blocks getting fatter. And the reason we want to know this is so we'll
        # know whether to update various workspace widgets.
        
        jobs = self.project.crunching_manager.jobs
        
        jobs_to_nodes = dict((job, job.node) for job in jobs)

                
        added_nodes = self.project.sync_crunchers()        
        # This is the heavy line here, which actually executes the Project's
        # `sync_crunchers` function.
        
        
        if any(
            (job not in jobs) or \
            (job.node.soft_get_block() is not old_node.soft_get_block())
            for job, old_node in jobs_to_nodes.iteritems()
               ):

            # What does this codition mean?
            # 
            # It means that there is at least one job that either:
            # (a) Was removed from the jobs list, or
            # (b) Changed the soft block it's pointing to.            
            #
            # The thing is, if there was a structural modification in the tree,
            # this condition must be True. Therefore we call this here:
            
            self.tree_structure_modified_at_unknown_location_emitter.emit()
            
            # Even though we are not sure that the tree structure was modified;
            # We have to play it safe. And since this condition doesn't happen
            # most of the time when crunching, we're not wasting too much
            # rendering time by assuming this is a structural modification.
            
            # Note that we didn't check if `added_nodes > 0`: This is because if
            # an `End` was added to the tree, it wouldn't have been counted in
            # `added_nodes`.
            
        elif added_nodes > 0:
            
            # If this condition is True, we know as a fact that there was no
            # structural modification, and we know as a fact that some blocks
            # have gotten fatter.
            
            self.tree_modified_at_unknown_location_emitter.emit()
            
        # todo: It would be hard but nice to know whether the tree changes were
        # on the path. This could save some rendering on SeekBar.
            
        return added_nodes

    
    def finalize_active_node(self):
        '''Finalize the changes made to the active node.'''
        self.active_node.finalize()
        
        self.active_node_finalized_emitter.emit()
        
        self.project.ensure_buffer(self.active_node, self.default_buffer)

        
    def __getstate__(self):
        my_dict = dict(self.__dict__)
        
        del my_dict['frame']
        del my_dict['timer_for_playing']
        del my_dict['simpack_grokker']
        del my_dict['simpack_wx_grokker']

        for (key, value) in my_dict.items():
            
            if isinstance(value, emitters.Emitter) or \
               isinstance(value, emitters.EmitterSystem):
                
                del my_dict[key]
            
        return my_dict

    
    def __setstate__(self, pickled_project):
        raise Exception
    


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the Emitter class.

See its documentation for more info.
'''

# todo: there should probably be some circularity check. Maybe actually
# circularity should be permitted?

# todo: make some way to emit from multiple emitters simulataneously, saving
# redundant calls to shared callable outputs.

import itertools
from garlicsim.general_misc import cute_iter_tools
from garlicsim.general_misc import misc_tools
        

class Emitter(object):
    '''
    An emitter you can `emit` from to call all its callable outputs.
    
    The emitter idea is a variation on the publisher-subscriber design pattern.

    Every emitter has a set of inputs and a set of outputs. The inputs, if there
    are any, must be emitters themselves. So when you `emit` on any of this
    emitter's inputs, it's as if you `emit`ted on this emitter as well.
    (Recursively, of course.)
    
    The outputs are a bit different. An emitter can have as outputs both (a)
    other emitters and (b) callable objects. (Which means, functions or
    function-like objects.)
    
    There's no need to explain (a): If `emitter_1` has as an output `emitter_2`,
    then `emitter_2` has as an input `emitter_1`, which works like how we
    explained above about inputs.
    
    But now (b): An emitter can have callables as outputs. (Without these, the
    emitter idea won't have much use.) These callables simply get called
    whenever the emitter or one of its inputs get `emit`ted.
    
    The callables that you register as outputs are functions that need to be
    called when the original event that caused the `emit` action happens.
    '''
        
    def __init__(self, inputs=(), outputs=(), name=None):
        '''
        Construct the emitter.
        
        `inputs` is a list of inputs, all of them must be emitters.
        
        `outputs` is a list of outputs, they must be either emitters or
        callables.
        
        `name` is a string name for the emitter.
        '''

        assert cute_iter_tools.is_iterable(inputs) and \
               cute_iter_tools.is_iterable(outputs)
        
        self._inputs = set()
        '''The emitter's inputs.'''
        
        self._outputs = set()
        '''The emitter's inputs.'''
        
        for output in outputs:
            self.add_output(output)
                        
        self.__total_callable_outputs_cache = None
        '''
        A cache of total callable outputs.
        
        This means the callable outputs of this emitter and any output emitters.
        '''
        
        self._recalculate_total_callable_outputs()        

        # We made sure to create the callable outputs cache before we add
        # inputs, so when we update their cache, it could use ours.
        for input in inputs:
            self.add_input(input)

        self.name = name
        '''The emitter's name.'''

    def get_inputs(self):
        '''Get the emitter's inputs.'''
        return self._inputs
    
    def get_outputs(self):
        '''Get the emitter's outputs.'''
        return self._outputs
                
    def _get_input_layers(self):
        '''
        Get the emitter's inputs as a list of layers.
        
        Every item in the list will be a list of emitters on that layer. For
        example, the first item will be a list of direct inputs of our emitter.
        The second item will be a list of *their* inputs. Etc.
        
        Every emitter can appear only once in this scheme: It would appear on
        the closest layer that it's on.
        '''

        input_layers = [self._inputs]
        current_layer = self._inputs
        while current_layer:
            
            next_layer = reduce(
                set.union,
                (input._inputs for input in current_layer),
                set()
            )
            
            for ancestor_layer in input_layers:
                assert isinstance(next_layer, set)
                next_layer -= ancestor_layer

            input_layers.append(next_layer)
            
            current_layer = next_layer        

        
        # assert sum(len(layer) for layer in input_layers) == \
        #        len(reduce(set.union, input_layers, set()))
            
        return input_layers
                
        
    def _recalculate_total_callable_outputs_recursively(self):
        '''
        Recalculate `__total_callable_outputs_cache` recursively.
        
        This will to do the recalculation for this emitter and all its inputs.
        '''
        
        # todo: I suspect this wouldn't work for the following case. `self` has
        # inputs `A` and `B`. `A` has input `B`. A callable output `func` was
        # just removed from `self`, so this function got called. We update the
        # cache here, then take the first input layer, which is `A` and `B` in
        # some order. Say `B` is first. Now, we do `recalculate` on `B`, but `A`
        # still got the cache with `func`, and `B` will take that. I need to
        # test this.
        # 
        # I have an idea how to solve it: In the getter of the cache, check the
        # cache exists, otherwise rebuild. The reason we didn't do it up to now
        # was to optimize for speed, but only `emit` needs to be fast and it
        # doesn't use the getter. We'll clear the caches of all inputs, and
        # they'll rebuild as they call each other.
        
        self._recalculate_total_callable_outputs()
        input_layers = self._get_input_layers()
        for input_layer in input_layers:
            for input in input_layer:
                input._recalculate_total_callable_outputs()
        
        
    def _recalculate_total_callable_outputs(self):
        '''
        Recalculate `__total_callable_outputs_cache` for this emitter.
        
        This will to do the recalculation for this emitter and all its inputs.
        '''
        children_callable_outputs = reduce(
            set.union,
            (emitter.get_total_callable_outputs() for emitter
             in self._get_emitter_outputs() if emitter is not self),
            set()
        )
        
        self.__total_callable_outputs_cache = \
            children_callable_outputs.union(self._get_callable_outputs())

    def add_input(self, emitter):
        '''Add an emitter as an input to this emitter.'''
        assert isinstance(emitter, Emitter)
        self._inputs.add(emitter)
        emitter._outputs.add(self)
        emitter._recalculate_total_callable_outputs_recursively()
        
    def remove_input(self, emitter):
        '''Remove an input from this emitter.'''
        assert isinstance(emitter, Emitter)
        self._inputs.remove(emitter)
        emitter._outputs.remove(self)
        emitter._recalculate_total_callable_outputs_recursively()
    
    def add_output(self, thing):
        '''Add an emitter or a callable as an output to this emitter.'''
        assert isinstance(thing, Emitter) or callable(thing)
        self._outputs.add(thing)
        if isinstance(thing, Emitter):
            thing._inputs.add(self)
        self._recalculate_total_callable_outputs_recursively()
        
    def remove_output(self, thing):
        '''Remove an output from this emitter.'''
        assert isinstance(thing, Emitter) or callable(thing)
        self._outputs.remove(thing)
        if isinstance(thing, Emitter):
            thing._inputs.remove(self)
        self._recalculate_total_callable_outputs_recursively()
        
    def disconnect_from_all(self): # todo: use the freeze here
        '''Disconnect the emitter from all its inputs and outputs.'''
        for input in self._inputs: 
            self.remove_input(input)
        for output in self._outputs:
            self.remove_output(output)
        
    def _get_callable_outputs(self):
        '''Get the direct callable outputs of this emitter.'''
        return set((
            output for output in self._outputs if callable(output)
        ))
    
    def _get_emitter_outputs(self):
        '''Get the direct emitter outputs of this emitter.'''
        return set((
            output for output in self._outputs if isinstance(output, Emitter)
        ))
        
    def get_total_callable_outputs(self):
        '''
        Get the total of callable outputs of this emitter.
        
        This means the direct callable outputs, and the callable outputs of
        emitter outputs.
        '''
        return self.__total_callable_outputs_cache
    
    def emit(self):
        '''
        Call all of the (direct or indirect) callable outputs of this emitter.
        
        This is the most important method of the emitter. When you `emit`, all
        the callable outputs get called in succession.
        '''
        # Note that this function gets called many times, so it should be
        # optimized for speed.
        for callable_output in self.__total_callable_outputs_cache:
            # We are using the cache directly instead of calling the getter, for
            # speed.
            callable_output()
    
    def __repr__(self):
        '''
        Get a string representation of the emitter.
        
        Example output:        
        <garlicsim_wx.general_misc.emitters.emitter.Emitter 'tree_modified' at
        0x1c013d0>
        '''
        return '<%s %sat %s>' % (
            misc_tools.shorten_class_address(
                       self.__class__.__module__,
                       self.__class__.__name__
                       ),
            ''.join(("'", self.name, "' ")) if self.name else '',
            hex(id(self))
        )
    """
    Unused:
    
    def _get_total_inputs(self):
        
        total_inputs_of_inputs = reduce(
            set.union,
            (emitter._get_total_inputs() for emitter
             in self._inputs if emitter is not self),
            set()
        )
        
        return total_inputs_of_inputs.union(self._inputs)
    """


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the Shell class.

See its documentation for more info.
'''

import wx.py.shell

from garlicsim_wx.widgets import WorkspaceWidget
import garlicsim
import garlicsim_wx


class Shell(wx.py.shell.Shell, WorkspaceWidget):
    '''
    A shell widget, allowing to execute arbitrary Python code.
    
    This is useful for examining the simulation.
    '''
    # todo: After every command in shell, must make the top emitter emit
    # todo: Make one namespace for the entire program.
    def __init__(self, frame):
        locals_for_shell = {
            'f': frame,
            'gp': frame.gui_project,
            'p': frame.gui_project.project,
            't': frame.gui_project.project.tree,
            'garlicsim': garlicsim,
            'garlicsim_wx': garlicsim_wx,
            'wx': wx,
        }
        wx.py.shell.Shell.__init__(self, frame, size=(100, 100),
                                   locals=locals_for_shell)
        WorkspaceWidget.__init__(self, frame)
        
        # Obscure: This causes the `site` module to add `help` and a few others
        # to `__builtin__`. For some reason `site` isn't imported when frozen
        # with py2exe, so here we make sure to import it.
        import site; del site
        
    



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines an EmitterSystem, which offers some benefits over Emitter.

See documentation of EmitterSystem for more info.
'''

from __future__ import with_statement

import itertools
from garlicsim.general_misc import cute_iter_tools

from emitter import Emitter


class FreezeCacheRebuildingContextManager(object):
    '''
    Context manager for freezing the cache rebuilding in an emitter system.
    
    When you do actions using this context manager, the emitters will not
    rebuild their cache when changing their inputs/outputs. When the outermost
    context manager has exited, all the caches for these emitters will get
    rebuilt.
    '''    
    def __init__(self, emitter_system):
        self.emitter_system = emitter_system
        
    def __enter__(self):
        assert self.emitter_system._cache_rebuilding_frozen >= 0
        self.emitter_system._cache_rebuilding_frozen += 1
        
    def __exit__(self, *args, **kwargs):
        self.emitter_system._cache_rebuilding_frozen -= 1
        if self.emitter_system._cache_rebuilding_frozen == 0:
            self.emitter_system._recalculate_all_cache()

            
class EmitterSystem(object):
    '''
    A system of emitters, representing a set of possible events in a program
    
    EmitterSystem offers a few advantages over using plain emitters.
    
    There are the `bottom_emitter` and `top_emitter`, which allow, respectively,
    to keep track of each `emit`ting that goes on, and to generate an `emit`ting
    that affects all emitters in the system.
    
    The EmitterSystem also offers a context manager, `.freeze_cache_rebuilding`.
    When you do actions using this context manager, the emitters will not
    rebuild their cache when changing their inputs/outputs. When the outermost
    context manager has exited, all the caches for these emitters will get
    rebuilt.
    '''
    # possible future idea: there is the idea of optimizing by cutting redundant
    # links between boxes. I'm a bit suspicious of it. The next logical step is
    # to make inputs and outputs abstract.
    def __init__(self):
        
        self._cache_rebuilding_frozen = 0
        self.freeze_cache_rebuilding = \
            FreezeCacheRebuildingContextManager(self)
        
        self.emitters = set()
        
        self.bottom_emitter = Emitter(self, name='bottom')
        self.emitters.add(self.bottom_emitter)
        
        self.top_emitter = Emitter(
            self,
            outputs=(self.bottom_emitter,),
            name='top',
        )
        self.emitters.add(self.top_emitter)
        
            
    def make_emitter(self, inputs=(), outputs=(), name=None):
        '''Create an emitter in this emitter system. Returns the emitter.'''

        # todo: allow one value in inputs and outputs. do in all emitter
        # constructors.
        
        inputs = set(inputs)
        inputs.add(self.top_emitter)
        outputs = set(outputs)
        outputs.add(self.bottom_emitter)
        emitter = Emitter(self, inputs, outputs, name)
        self.emitters.add(emitter)
        return emitter
    
    def remove_emitter(self, emitter):
        '''
        Remove an emitter from this system, disconnecting it from everything.
        '''
        with self.freeze_cache_rebuilding:
            emitter.disconnect_from_all()
        self.emitters.remove(emitter)
        
    def _recalculate_all_cache(self):
        '''Recalculate the cache for all the emitters.'''
        self.bottom_emitter._recalculate_total_callable_outputs_recursively()
    




# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the CuteTimer class.

See its documentation for more information.
'''

import wx

from garlicsim_wx.general_misc import cute_base_timer


__all__ = ['CuteTimer']


class CuteTimer(wx.Timer, cute_base_timer.CuteBaseTimer):
    '''A wx.Timer that allows central stopping.'''
    # todo: Should use PyTimer?
    def __init__(self, parent=None, id=wx.ID_ANY):
        wx.Timer.__init__(self, parent, id)
        cute_base_timer.CuteBaseTimer.__init__(self, parent)
        


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the CuteDialog class.

See its documentation for more info.
'''

import wx

class CuteDialog(wx.Dialog):
    '''Dialog that works around a Mac bug by centering itself on the screen.'''
    def ShowModal(self):
        if wx.Platform == '__WXMAC__':
            wx.CallAfter(lambda: self.Centre(wx.BOTH))
        return super(CuteDialog, self).ShowModal()
    


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the WorkspaceMenu class.

See its documentation for more info.
'''

import wx

from garlicsim_wx.general_misc.cute_menu import CuteMenu

class WorkspaceMenu(CuteMenu):
    '''Menu for manipulating the workspace.'''
    def __init__(self, frame):
        super(WorkspaceMenu, self).__init__()
        self.frame = frame
        self._build()
        
        
    def _build(self):

        frame = self.frame
        
        self.save_workspace_button = self.Append(
            -1,
            '&Save workspace...',
            ''' Save the current workspace configuration, so that it may be \
recalled in the future'''
        )
        self.save_workspace_button.Enable(False)
        
        
        self.delete_workspace_button = self.Append(
            -1,
            '&Delete workspace...',
            ' Delete one of the saved workspace configurations'
        )
        self.delete_workspace_button.Enable(False)
        
        
        self.AppendSeparator()
        
                
        self.delete_workspace_button = self.Append(
            -1,
            'De&fault workspace',
            ' Use the factory-default workspace configuration'
        )
        self.delete_workspace_button.Enable(False)
                



# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
Defines the CuteMenu class.

See its documentation for more information.
'''

import wx
from garlicsim.general_misc.third_party import abc


class CuteMenu(wx.Menu):
    '''Menu class that allows easy adding of menus.'''
    
    __metaclass__ = abc.ABCMeta
        
    @abc.abstractmethod
    def _build(self):
        '''Build the menu, populating it with items and/or submenus.'''
    
    @staticmethod
    def add_menus(menus):
        '''
        Build a menu from a sequence of smaller menus.
        
        A separator will come between the items of one menu to the the items of
        the next.
        '''
        big_menu = UnbuildableCuteMenu()
    
        first_run = True
    
        for menu in menus:

            assert isinstance(menu, CuteMenu)
            
            if not first_run:
                big_menu.AppendSeparator()
                assert big_menu.frame is menu.frame
            else:
                big_menu.frame = menu.frame
                first_run = False
                
            type(menu).__dict__['_build'](big_menu)
                
        return big_menu
    
    
class UnbuildableCuteMenu(CuteMenu):
    '''
    CuteMenu that can't be built.
    
    This is useful when creating a Menu with `CuteMenu.add_menus`.
    '''
    def _build(self):
        raise Exception("Can't _build an UnbuildableCuteMenu.")


# Copyright 2009-2010 Ram Rachum. No part of this program may be used, copied
# or distributed without explicit written permission from Ram Rachum.

'''
This module defines the ThreadTimer class.

See its documentation for more info.
'''
# todo: daemonize?
# todo: kickass idea: make all timers use one thread that will sleep smartly
# to send all events correctly.

import threading
import time

import wx

from garlicsim_wx.general_misc import cute_base_timer


wxEVT_THREAD_TIMER = wx.NewEventType()
EVT_THREAD_TIMER = wx.PyEventBinder(wxEVT_THREAD_TIMER, 1)


class ThreadTimer(cute_base_timer.CuteBaseTimer):
   '''
   A timer for a wxPython app which runs on a different thread.
   
   This solved a problem of wxPython timers being late when the program is
   busy.
   '''
   
   n = 0
   '''The number of created thread timers.'''
   
   
   def __init__(self, parent):
      '''
      Construct the ThreadTimer.
      
      `parent` is the parent window.
      '''
      
      cute_base_timer.CuteBaseTimer.__init__(self, parent)
      
      self.parent = parent
      '''The parent window.'''
      
      ThreadTimer.n += 1
      self.wx_id = wx.NewId()
      '''The ID of this timer, given by wxPython.'''
      
      self.__init_thread()
      self.alive = False
      '''Flag saying whether this timer is running.'''

   def __init_thread(self):
      '''Create the thread.'''
      thread_name = ''.join(('Thread used by ThreadTimer no. ', str(self.n)))
      self.thread = Thread(self, name=thread_name)
      # Overwriting previous thread, so it'll get garbage-collected, hopefully
   
   def start(self, interval):
      '''Start the timer.'''
      if self.alive:
         self.stop()
      self.interval = interval
      self.alive = True
      self.thread.start()

   def stop(self):
      '''Stop the timer.'''
      self.alive = False
      self.thread.retired = True
      self.__init_thread()
   
   # Crutch for compatibilty with wx.Timer:
   Start = start
   Stop = stop
      
   def GetId(self):
      '''Get the wx ID of this timer.'''
      return self.wx_id

      
class Thread(threading.Thread):
   '''Thread used as a timer for wxPython programs.'''
   def __init__(self, parent, name):
      threading.Thread.__init__(self, name=name)
      self.parent = parent
      self.retired = False
      
   def run(self):
      '''Run the thread. Internal function.'''
      interval_in_seconds = self.parent.interval / 1000.0
      def sleep():
            time.sleep(interval_in_seconds)

      sleep()
      try:
         while self.parent.alive is True and self.retired is False:
            event = wx.PyEvent(self.parent.wx_id)
            event.SetEventType(wxEVT_THREAD_TIMER)
            wx.PostEvent(self.parent.parent, event)
            sleep()
      except:
         return # Just so it wouldn't raise an error when wx is shutting down